#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan  2 15:03:42 2019

@author: user
"""
import time
import itertools
import os
import tables
import shutil
from glob import glob

import fastparquet as pq
import numpy as np
import pandas as pd
import psycopg2 as pg

import grimsel
import grimsel.auxiliary.sqlutils.aux_sql_func as aql
import grimsel.core.autocomplete as ac
import grimsel.core.table_struct as table_struct
from grimsel import _get_logger

logger = _get_logger(__name__)


FORMAT_RUN_ID = '{:04d}'  # modify for > 9999 model runs


class _HDFWriter:
    ''' Mixing class for :class:`CompIO` and :class:`DataReader`. '''

    def write_hdf(self, tb, df, put_append):
        '''
        Opens connection to HDF file and writes output.

        Parameters
        ----------
        put_append: str, one of `('append', 'put')`
            Write one-time table to the output file (`'put'`) or append
            to existing table (`'append'`).

        '''

        with pd.HDFStore(self.cl_out, mode='a') as store:

            method_put_append = getattr(store, put_append)
            method_put_append(tb, df, data_columns=True, format='table',
                              complevel=9, complib='blosc:blosclz')

class _ParqWriter:
    ''' Mixing class for :class:`CompIO` and :class:`DataReader`. '''

    def write_parquet(self, fn, df, engine):
        '''
        Opens connection to HDF file and writes output.

        Parameters
        ----------
        fn: str
            filename for table writing
        df: pandas DataFrame
            table to be written
        engine: str
            engine name as in the pandas DataFrame.to_parquet parameter
        '''

        if self.output_target == 'fastparquet':
            df.to_parquet(fn, engine='fastparquet',
                          compression='gzip',)

#            if 'run_id' in df.columns:
#                df.to_parquet(fn, #append=os.path.isfile(fn),
#                              engine='fastparquet',
#                              compression='gzip',
#                              partition_cols=['run_id'])
#            else:
#                df.to_parquet(fn, #append=os.path.isfile(fn),
#                              engine='fastparquet',
#                              compression='gzip'
#                              )


        else:
            raise RuntimeError('Writing using parquet engine %s '
                               'not implemented.'%self.output_target)



class CompIO(_HDFWriter, _ParqWriter):
    '''
    A CompIO instance takes care of extracting a single variable/parameter from
    the model and of writing a single table to the database.
    '''

    def __init__(self, tb, cl_out, comp_obj, idx, connect, output_target,
                 model=None):

        self.tb = tb
        self.cl_out = cl_out
        self.comp_obj = comp_obj
        self.output_target = output_target
        self.connect = connect
        self.model = model

        self.columns = None  # set in index setter
        self.run_id = None  # set in call to self.write_run

        self.index = tuple(idx) if not isinstance(idx, tuple) else idx

        self.coldict = aql.get_coldict()

    def post_processing(self, df):
        ''' Child-specific method called after reading. '''
        return df

    def to_df(self):
        '''
        Calls classmethods _to_df.

        Is overwritten in DualIO, where _to_df is not used as classmethod.

        '''

        return self._to_df(self.comp_obj,
                           [c for c in self.index if not c == 'bool_out'])

    def init_output_table(self):
        '''
        Initialization of output table.

        Calls the :func:`aux_sql_func` method with appropriate parameters.

        .. note:
           Keys need to be added in post-processing due to table
           writing performance.

        '''

        logger.info('Generating output table {}'.format(self.tb))
        col_names = self.index + ('value',)
        cols = [(c,) + (self.coldict[c][0],) for c in col_names]
        cols += [('run_id', 'SMALLINT')]
        pk = []  # pk added later for writing/appending performance
        unique = []

        aql.init_table(tb_name=self.tb, cols=cols,
                       schema=self.cl_out,
                       ref_schema=self.cl_out, pk=pk,
                       unique=unique, bool_auto_fk=False, db=self.connect.db,
                       con_cur=self.connect.get_pg_con_cur())


    def _to_file(self, df, tb):
        '''
        Casts the data types of the output table and writes the
        table to the output HDF file.

        '''

        dtype_dict = {'value': np.dtype('float64'),
                      'bool_out': np.dtype('bool')}
        dtype_dict.update({col: np.dtype('int32') for col in df.columns
                           if not col in ('value', 'bool_out')})

        df = df.astype({col: dtype for col, dtype in dtype_dict.items()
                        if col in df.columns})


        if self.output_target == 'hdf5':
            self.write_hdf(tb, df, 'append')

        elif self.output_target in ['fastparquet']:

            fn = os.path.join(self.cl_out,
                              tb + ('_%s'%FORMAT_RUN_ID).format(self.run_id) + '.parq')

            self.write_parquet(fn, df, engine=self.output_target)

        else:
            raise RuntimeError('_to_file: no '
                               'output_target applicable')


    def _to_sql(self, df, tb):

        df.to_sql(tb, self.connect.get_sqlalchemy_engine(),
                  schema=self.cl_out, if_exists='append', index=False)

    def _finalize(self, df, tb=None):
        ''' Add run_id column and write to database table '''

        tb = self.tb if not tb else tb
        logger.info('Writing {} to {}.{}'.format(self.comp_obj.name,
                                                 self.cl_out, tb))

        # value always positive, directionalities expressed through bool_out
        df['value'] = df['value'].abs()

        df['run_id'] = self.run_id

        t = time.time()

        if self.output_target in ['hdf5', 'fastparquet']:
            self._to_file(df, tb)
        elif self.output_target == 'psql':
            self._to_sql(df, tb)
        else:
            raise RuntimeError('_finalize: no '
                               'output_target applicable')

        logger.info(' ... done in %.3f sec'%(time.time() - t))

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, value):
        ''' Makes sure idx is tuple and updates columns attribute. '''
        self._index = (value,) if not isinstance(value, tuple) else value
        self.columns = list(self.index + ('value',))
        self.columns = [c for c in self.columns if not c == 'bool_out']

    def get_df(self):

        df = self.to_df()
        df = self.post_processing(df)

        return df

    def write(self, run_id):

        self.run_id = run_id

        df = self.get_df()

        self._finalize(df)

    def _node_to_plant(self, pt):
        '''
        TODO: THIS SHOULD BE IN MAPS!!!!
        TODO: THIS SHOULD ALSO INCLUDE ca_id FOR DMND!
        Method for translation of node_id to respective plant_id
        in the cases of demand and inter-node transmission. This is used to
        append demand/inter-nodal transmission to pwr table.
        Returns a dictionary node -> plant
        Keyword arguments:

        * pt -- string, selected plant type for translation
        '''
        df_np = self.model.df_def_plant[['nd_id', 'pp_id', 'pp', 'pt_id']]
        df_pt = self.model.df_def_pp_type[['pt_id', 'pt']]
        mask_pt = df_pt.pt.apply(lambda x: x.replace('_ST', '')) == pt
        slct_pp_type = df_pt.loc[mask_pt, 'pt_id'].astype(int).iloc[0]

        mask_tech = df_np['pt_id'] == slct_pp_type
        df_np_slct = df_np.loc[mask_tech]
        dict_node_plant_slct = df_np_slct.set_index('nd_id')
        dict_node_plant_slct = dict_node_plant_slct['pp_id'].to_dict()
        return dict_node_plant_slct

    def __repr__(self):

        return 'Comp_obj: ' + str(self.comp_obj)


class DualIO(CompIO):
    '''
    Base class for dual values. Performs the data extraction of constraint
    shadow prices.
    '''

    def to_df(self):

        dat = [ico + (self.model.dual[self.comp_obj[ico]],)
               for ico in self.comp_obj
               if self.comp_obj[ico].active]
        return pd.DataFrame(dat, columns=self.columns)


class VariabIO(CompIO):
    '''
    Base class for variables. Performs the data extraction of variable
    objects.
    Also: Special methods related to the handling of negative plant
    variables defined by setlst[sll] + setlst[curt]
    as well as storage charging.

    '''

    @classmethod
    def _to_df(cls, obj, cols):
        ''' Converts pyomo variable to DataFrame.'''

        df = pd.Series(obj.extract_values()).fillna(0).reset_index()
        df.columns = list(cols) + ['value']

        return df

    def post_processing(self, df):
        '''
        Calls _set_bool_out prior to writing.
        Input arguments:
        * df -- DataFrame; primary dataframe
        '''

        return self._set_bool_out(df) if 'bool_out' in self.index else df

    def _set_bool_out(self, df):
        '''
        Set the bool_out values according to the pp_id.

        pp_ids corresponding to curtailment, charging, and sales have
        bool_out=True.
        * df -- DataFrame; primary dataframe
        '''

        if self.comp_obj.name in ['pwr_st_ch', 'erg_ch_yr']:
            df['bool_out'] = True
        else:
            df['bool_out'] = False

            pp_true = self.model.setlst['curt'] + self.model.setlst['sll']
            df.loc[df.pp_id.isin(pp_true), 'bool_out'] = True

        return df


class ParamIO(CompIO):
    '''
    Base class for parameters.

    Is inherited by :class:`DmndIO`.

    Only contains the parameter ``_to_df`` classmethod.

    '''

    @classmethod
    def _to_df(cls, obj, cols):
        ''' Converts pyomo parameter to DataFrame. '''

        df = pd.Series(obj.extract_values()).fillna(0).reset_index()
        if df.empty:
            df = pd.DataFrame(columns=list(cols) + ['value'])
        else:
            if not cols and len(df) is 1:
                df = df[[0]].rename(columns={0: 'value'})
            else:
                df.columns = list(cols) + ['value']

        return df


class TransmIO(VariabIO):
    """
    Special methods related to the translation of nodes to
    transmission plant names and
    to the simplified representation after aggregating secondary nodes.
    """

    def post_processing(self, df):
        ''' Write aggregated transmission table to pwr. '''

        dfagg = self.aggregate_nd2(df)
        dfagg = self._translate_trm(dfagg)

        self._finalize(dfagg, 'var_sy_pwr')

        return self.add_bool_out_col(df)

    def aggregate_nd2(self, dfall):
        '''
        Aggregates trm table over all secondary nodes for simplification and
        to append to the pwr table.
        '''
        # mirror table to get both directions
        dfall = pd.concat([dfall,
                           dfall.assign(nd_2_id=dfall.nd_id,
                                        nd_id=dfall.nd_2_id,
                                        value=-dfall.value)])

        dict_nhours = {nd_id:
                       self.model._tm_objs[self.model.dict_nd_tm_id[nd_id]].nhours
                       for nd_id in dfall.nd_id.unique()}

        def avg_to_nhours(x):

            if self.model.is_min_node[x.name]:
                return x.reset_index()
            else:  # reduce time resolution
                nhours = dict_nhours[x.name[0]]
                nhours_2 = dict_nhours[x.nd_2_id.iloc[0]]

                x['sy'] = np.repeat(np.arange(
                                    np.ceil(len(x) / (nhours / nhours_2))),
                                    nhours / nhours_2)

                idx = [c for c in x.columns if not c == 'value']
                x = x.pivot_table(index=idx, values='value', aggfunc=np.mean)
                return x.reset_index()

        dfall = (dfall.groupby(['nd_id', 'nd_2_id'], as_index=True)
                      .apply(avg_to_nhours)
                      .reset_index(drop=True))

        dfexp = dfall.loc[dfall.value > 0]
        dfexp = dfexp.groupby(['sy', 'nd_id', 'ca_id'])['value'].sum()
        dfexp = dfexp.reset_index()
        dfexp['bool_out'] = True

        dfimp = dfall.loc[dfall.value < 0]
        dfimp = dfimp.groupby(['sy', 'nd_id', 'ca_id'])['value'].sum()
        dfimp = dfimp.reset_index()
        dfimp['bool_out'] = False

        dfagg = pd.concat([dfexp, dfimp], axis=0)

        dict_nd_weight = {key: self.model.nd_weight[key].value
                          for key in self.model.nd_weight}
        dfagg['value'] /= dfagg.nd_id.replace(dict_nd_weight)

        return dfagg

    def _translate_trm(self, df):

        df['pp_id'] = df.nd_id.replace(self._node_to_plant('TRNS'))
        df.drop('nd_id', axis=1, inplace=True)

        return df

    def add_bool_out_col(self, df):
        ''' The bool out column value depends on the sign of the data. '''

        df['bool_out'] = False
        df.loc[df.value < 0, 'bool_out'] = True

        return df


class DmndIO(ParamIO):
    '''
    Demand is appended to the *pwr* table after translating the nd_id to
    the corresponding "power plant" pp_id.

    '''

    def post_processing(self, df):

        dfpp = self._translate_dmnd(df.copy())
        dfpp['bool_out'] = True
        dfpp = dfpp[['sy', 'pp_id', 'ca_id', 'value', 'bool_out']]

        self._finalize(dfpp, 'var_sy_pwr')

        return df

    def _translate_dmnd(self, df):
        '''
        Translate the demand ``pf_id`` to the corresponding ``pp_id``s

        This is based on a mapping ``pf_id`` |rarr| ``nd_id`` |rarr| ``pp_id``.
        The ``pp_id`` definition for demand is retrieved from the
        ``ModelBase.df_def_plant`` table.
        '''

        dict_ndpp = self._node_to_plant('DMND')
        df['pp_id'] = df.nd_id.replace(dict_ndpp)

        return df


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def skip_if_resume_loop(f):
    def wrapper(self, *args, **kwargs):
        if self.resume_loop:
            pass
        else:
            f(self, *args, **kwargs)
    return wrapper


def skip_if_no_output(f):
    def wrapper(self, *args, **kwargs):
        if self.no_output:
            pass
        else:
            f(self, *args, **kwargs)
    return wrapper


class ModelWriter():
    '''
    The IO singleton class manages the TableIO instances and communicates with
    other classes. Manages database connection.
    '''

    io_class_dict = {'var': VariabIO,
                     'var_tr': TransmIO,
                     'par_dmnd': DmndIO,
                     'par': ParamIO,
                     'dual': DualIO}

    _default_init = {'sc_warmstart': False,
                     'resume_loop': False,
                     'replace_runs_if_exist': False,
                     'model': None,
                     'output_target': 'hdf5',
                     'sql_connector': None,
                     'no_output': False,
                     'dev_mode': False,
                     'coll_out': None,
                     'keep': None,
                     'drop': None,
                     'db': None}

    def __init__(self, **kwargs):
        """
        """

        self.run_id = None  # set in call to self.write_run
        self.dict_comp_obj = {}


        # define instance attributes and update with kwargs
        for key, val in self._default_init.items():
            setattr(self, key, val)
        self.__dict__.update(kwargs)

        self.dict_comp_idx = None
        self.dict_comp_table = None
        self.dict_comp_group = None

        ls = 'Output collection: {}; resume loop={}'
        logger.info(ls.format(self.cl_out, self.resume_loop))

        self.reset_tablecollection()


    def _make_table_dicts(self, keep=None, drop=None):
        '''
        Get the dictionaries describing all tables.

        Also used in the class method ``post_process_index``,
        therefore module function.
        '''

        keep = list(table_struct.DICT_COMP_IDX) if not keep else keep
        keep = set(keep) - set(drop if drop else [])

        options = list(table_struct.DICT_COMP_IDX)
        unknowns = [tb for tb in keep if not tb in options]
        if unknowns:
            raise RuntimeError(('Unknown table selection %s. Possible options '
                               'are %s')%(str(unknowns), str(options)))

        filter_dict = lambda d: {k: v for k, v in d.items() if k in keep}

        self.dict_comp_idx = filter_dict(table_struct.DICT_COMP_IDX)
        self.dict_comp_table = filter_dict(table_struct.DICT_COMP_TABLE)
        self.dict_comp_group = filter_dict(table_struct.DICT_COMP_GROUP)

    def reset_tablecollection(self):
        '''
        Reset the SQL schema or hdf file for model output writing.

        '''

        if self.output_target == 'psql':
            self._reset_schema()
        elif self.output_target == 'hdf5':
            self._reset_hdf_file()
        elif self.output_target in ['fastparquet']:
            self._reset_parquet_file()

    @skip_if_resume_loop
    def _reset_hdf_file(self):

        ModelWriter.reset_hdf_file(self.cl_out, not self.dev_mode)

    def _reset_parquet_file(self):

        ModelWriter.reset_parquet_file(self.cl_out, not self.dev_mode,
                                       self.resume_loop)

    @staticmethod
    def reset_hdf_file(fn, warn):
        '''
        Deletes existing hdf5 file and creates empty one.

        Parameters
        ----------
        fn: str
            filename
        warn: bool
            prompt user input if the file exists

        '''
#        pass

        if os.path.isfile(fn):

            try:
                max_run_id = pd.read_hdf(fn, 'def_run',
                                         columns=['run_id']).run_id.max()
            except Exception as e:
                logger.error(e)
                logger.warn('reset_hdf_file: Could not determine max_run_id '
                            '... setting to None.')
                max_run_id = None

            if warn:
                input(
'''
~~~~~~~~~~~~~~~   WARNING:  ~~~~~~~~~~~~~~~~
You are about to delete existing file {fn}.
The maximum run_id is {max_run_id}.

Hit enter to proceed.
'''.format(fn=fn, max_run_id=max_run_id)
)

            logger.info('Dropping output file {}'.format(fn))
            os.remove(fn)

    def reset_parquet_file(dirc, warn, resume_loop, ):
        '''
        Deletes existing parquet file folder and creates empty one.

        Parameters
        ----------
        dirc: str
            parquet directory name
        warn: bool
            prompt user input if the file exists

        '''

        if os.path.isdir(dirc) and not resume_loop:

            try:
                max_run_id = pd.read_parquet(os.path.join(dirc, 'def_run.parq'),
                                         columns=['run_id']).run_id.max()

            except Exception as e:
                logger.error(e)
                logger.warn('reset_parquet_file: Could not determine max_run_id '
                            '... setting to None.')
                max_run_id = None

            if warn:
                input(
'''
~~~~~~~~~~~~~~~   WARNING:  ~~~~~~~~~~~~~~~~
You are about to delete existing directory {dirc}.
The maximum run_id is {max_run_id}.

Hit enter to proceed.
'''.format(dirc=dirc, max_run_id=max_run_id)
)

            logger.info('Dropping parquet output directory {}'.format(dirc))
            shutil.rmtree(dirc)

        elif os.path.isdir(dirc) and resume_loop:

            logger.info('Deleting run_ids >= resume_loop = '
                        '{:d}'.format(resume_loop))

            del_fn = [fn for fn in glob(os.path.join(dirc, '*[0-9].parq')) if
                      int(fn.split('_')[-1].replace('.parq', ''))
                      >= resume_loop]

            if del_fn:
                for fn in del_fn:
                    logger.info('... deleting file {}'.format(fn))
                    os.remove(fn)

                fn_run = os.path.join(dirc, 'def_run.parq')
                df_def_run = pd.read_parquet(fn_run)
                df_def_run = df_def_run.query('run_id < %d'%resume_loop)
                df_def_run = df_def_run.reset_index(drop=True)
                pq.write(fn_run, df_def_run, append=False, compression='GZIP')

            else:
                logger.info('... nothing to delete.')

        if not os.path.isdir(dirc):

            os.mkdir(dirc)

    @skip_if_resume_loop
    def _reset_schema(self):

        aql.reset_schema(self.cl_out, self.sql_connector.db,
                         not self.dev_mode)

    def init_output_schema(self):

        aql.exec_sql('CREATE SCHEMA IF NOT EXISTS ' + self.cl_out,
                     db=self.db, )

    @skip_if_no_output
    def init_compio_objs(self):
        '''
        Initialize all output table IO objects.
        '''

        self._make_table_dicts(keep=self.keep, drop=self.drop)

        if __name__ == '__main__':
            comp, idx = 'pwr_st_ch', self.dict_comp_idx['pwr_st_ch']
        for comp, idx in self.dict_comp_idx.items():
            if not hasattr(self.model, comp):
                logger.warning(('Component {} does not exist... '
                               'skipping init CompIO.').format(comp))
            else:
                logger.debug('Adding component %s to dict_comp_obj'%comp)
                comp_obj = getattr(self.model, comp)

                grp = self.dict_comp_group[comp]
                if self.dict_comp_table[comp] in self.io_class_dict:
                    io_class = self.io_class_dict[self.dict_comp_table[comp]]
                elif grp in self.io_class_dict:
                    io_class = self.io_class_dict[self.dict_comp_group[comp]]
                else:
                    io_class = self.io_class_dict[self.dict_comp_group[comp].split('_')[0]]

                io_class_kwars = dict(tb=self.dict_comp_table[comp],
                                      cl_out=self.cl_out,
                                      comp_obj=comp_obj,
                                      idx=idx,
                                      connect=self.sql_connector,
                                      output_target=self.output_target,
                                      model=self.model)

                self.dict_comp_obj[comp] = io_class(**io_class_kwars)

    @skip_if_no_output
    def write_all(self):

        ''' Calls the write methods of all CompIO objects. '''

        for comp, io_obj in self.dict_comp_obj.items():

            io_obj.write(self.run_id)

    @skip_if_no_output
    def init_all(self):
        '''
        Initializes all SQL tables.

        Calls the init_output_table methods of all CompIO instances.
        '''

        if self.output_target == 'psql':

            coldict = aql.get_coldict(self.cl_out, self.sql_connector.db)

            for comp, io_obj in self.dict_comp_obj.items():

                io_obj.coldict = coldict
                io_obj.init_output_table()

        elif self.output_target in ['hdf5', 'fastparquet']:

            pass

    def delete_run_id(self, run_id=False, operator='>='):
        '''
        In output tables delete all rows with run_id >=/== the selected value.

        Used in :
            1. in ModelLoop.perform_model_run if replace_runs_if_exist == True
                with operator '=' to remove current run_id
                from all tables prior to writing

        TODO: The SQL part would be better fit with the aux_sql_func module.
        '''

        if run_id:



            # Get overview of all tables
            list_all_tb_0 = [list(itb_list + '_' + itb[0] for itb
                                  in getattr(table_struct, itb_list)
                                  if not len(itb) == 3)
                             for itb_list in table_struct.list_collect]
            self.list_all_tb = list(itertools.chain(*list_all_tb_0))
            self.list_all_tb += ['def_run']

            for itb in self.list_all_tb:

                if self.output_target == 'fastparquet':

                    self._delete_run_id_parquet(tb=itb, run_id=run_id)

                elif self.output_target == 'psql':

                    logger.info('Deleting from ' + self.cl_out + '.' + itb
                                + ' where run_id {} {}'.format(operator,
                                                               str(run_id)))
                    exec_strg = '''
                                DELETE FROM {cl_out}.{tb}
                                WHERE run_id {op} {run_id};
                                '''.format(cl_out=self.cl_out, tb=itb,
                                           run_id=run_id, op=operator)
                    try:
                        aql.exec_sql(exec_strg, db=self.db)
                    except pg.ProgrammingError as e:
                        logger.error(e)
                        raise(e)

    def _delete_run_id_parquet(self, tb, run_id):

        pat = os.path.join(self.cl_out, ('{}_%s.*'%FORMAT_RUN_ID).format(tb, run_id))
        fn_del = glob(pat)

        try:
            assert fn_del, 'Pattern not found: %s.'%pat
            assert not len(fn_del) > 1, \
                        'Found more than one table to delete: %s.'%fn_del

            os.remove(fn_del[0])

            logger.info('Successfully deleted table '
                        '{} of model run {:d}'.format(tb, run_id))
        except Exception as e:
            logger.error(e)


#    @classmethod
#    def post_process_index(cls, sc, db, drop=False):
#
#        coldict = aql.get_coldict(sc, db)
#
#        dict_idx, dict_table, _ = ModelWriter.get_table_dicts()
#
#        list_tables = aql.get_sql_tables(sc, db)
#
#        for comp, index in dict_idx.items():
#
#            if not dict_table[comp] in list_tables:
#                logger.warning('Table ' + comp + ' does not exist... skipping '
#                      'index generation.')
#            else:
#
#                tb_name = dict_table[comp]
#
#                logger.info('tb_name:', tb_name)
#
#                pk_list = index + ('run_id',)
#
#                fk_dict = {}
#                for c in pk_list:
#                    if len(coldict[c]) > 1:
#                        fk_dict[c] = coldict[c][1]
#
#
#                pk_kws = {'pk_list': ', '.join(pk_list),
#                          'tb': tb_name, 'cl_out': sc}
#                exec_str = ('''
#                            ALTER TABLE {cl_out}.{tb}
#                            DROP CONSTRAINT IF EXISTS {tb}_pkey;
#                            ''').format(**pk_kws)
#                if not drop:
#                    exec_str += ('''
#                                 ALTER TABLE {cl_out}.{tb}
#                                 ADD CONSTRAINT {tb}_pkey
#                                 PRIMARY KEY ({pk_list})
#                                 ''').format(**pk_kws)
#                logger.debug(exec_str)
#                aql.exec_sql(exec_str, db=db)
#
#                for fk_keys, fk_vals in fk_dict.items():
#                    fk_kws = {'cl_out': sc, 'tb': tb_name,
#                              'fk': fk_keys, 'ref': fk_vals}
#
#                    exec_str = ('''
#                                ALTER TABLE {cl_out}.{tb}
#                                DROP CONSTRAINT IF EXISTS fk_{tb}_{fk};
#                                ''').format(**fk_kws)
#
#                    if not drop:
#                        exec_str += ('''
#                                     ALTER TABLE {cl_out}.{tb}
#                                     ADD CONSTRAINT fk_{tb}_{fk}
#                                     FOREIGN KEY ({fk})
#                                     REFERENCES {ref}
#                                     ''').format(**fk_kws)
#                    logger.debug(exec_str)
#                    aql.exec_sql(exec_str, db=db)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class TableReader():
    '''
    Reads tables from input data sources and makes them attributes of the
    model attribute.
    '''

    def __init__(self, sql_connector, sc_inp, data_path, model):

        self.sqlc = sql_connector
        self.sc_inp = sc_inp
        self.data_path = (data_path if
                          isinstance(data_path, (tuple, list))
                          else [data_path])
        self.model = model

        if not self.sc_inp and not self.data_path:
            logger.warning('Falling back to grimsel default csv tables.')
            self.data_path = os.path.abspath(os.path.join(grimsel.__path__[0],
                                                          '..', 'input_data'))


        self.table_set, self.dict_tb_path = self._get_table_dict()

    def _get_table_dict(self):
        '''
        Obtain list of tables in the relevant data source.

        TODO: Update PSQL

        '''

        if self.sc_inp:
            return aql.get_sql_tables(self.sc_inp, self.sqlc.db)

        elif self.data_path:

            # path -> tables list
            dict_pt_tb = {path: [fn.replace('.csv', '')
                          for fn in next(os.walk(path))[-1]]
                          for path in self.data_path}

            table_set = set(itertools.chain.from_iterable(
                                                        dict_pt_tb.values()))

            # table -> under which paths
            dict_tb_path = {tb: [pt for pt, tb_list in dict_pt_tb.items() if
                            tb in tb_list] for tb in table_set}

            return table_set, dict_tb_path

    def _expand_table_families(self, dct):
        '''
        Searches for tables with identical name + suffix.
        Updates the dct.
        '''

        dct_add = {}
        for table, filt in dct.items():

            tbs_other = [tb for tb in self.table_set
                         if table in tb and not tb == table]

            if tbs_other:
                dct_add.update({tb: filt for tb in tbs_other})

        dct.update(dct_add)

    def df_from_dict(self, dct):
        '''
        Reads filtered input tables and assigns them to instance
        attributes.
        '''

        self._expand_table_families(dct)

        for table, filt in dct.items():

            list_df, tb_exists, source_str = self.get_input_table(table, filt)

            df = pd.concat(list_df, axis=0, sort=False) if tb_exists else None

            setattr(self.model, 'df_' + table, df)

            if not tb_exists:
                warn_str = ('Input table {tb} does not exist. Setting model '
                            'attribute df_{tb} to None.')
                logger.warning(warn_str.format(tb=table))
            else:
                filt = ('filtered by ' if len(filt) > 0 else '') +\
                   ', '.join([str(vvv[0]) + ' in ' + str(vvv[1]) for vvv in filt
                              if not len(vvv[1]) == 0])
                logger.info(('Reading input table {tb} {flt} from '
                       '{source_str}').format(tb=table, flt=filt,
                                              source_str=source_str))

    def get_input_table(self, table, filt):
        '''
        Returns list of tables.
        '''

        if self.sc_inp:

            tb_exists = table in aql.get_sql_tables(self.sc_inp, self.sqlc.db)

            if tb_exists:
                list_df = [aql.read_sql(self.sqlc.db, self.sc_inp,
                                        table, filt)]
            source = '%s %s.%s'%(self.sqlc.db, self.sc_inp, table)

        else:
            list_df = []

            tb_exists = table in self.dict_tb_path

            paths = self.dict_tb_path[table] if tb_exists else []

            source = []

            for path in paths:

                fn = os.path.join(path, '{}.csv'.format(table))
                source.append(fn)

                df = pd.read_csv(fn)

                logger.debug('Done reading, filtering according to {}'.format(filt))

                for col, vals in filt:
                    if isinstance(col, str):  # single column filtering
                        mask = df[col].isin(vals)
                    elif isinstance(col, (list, tuple)):  # multiple columns
                        mask = df[list(col)].apply(tuple, axis=1).isin(vals)
                    df = df.loc[mask]

                list_df.append(df)

        return ((list_df if tb_exists else None), tb_exists,
                (' from {}'.format(' and '.join(source)) if tb_exists else ''))

class DataReader(_HDFWriter, _ParqWriter):
    '''

    '''

    runtime_tables = [('tm_soy', ['sy', 'tm_id']),
                      ('hoy_soy', ['hy', 'tm_id']),
                      ('tm_soy_full', ['sy', 'tm_id']),
                      ('sy_min_all', ['sy_min', 'tm_id']),
                      ]

    def __init__(self, **kwargs):

        defaults = {'resume_loop': False,
                    'replace_runs_if_exist': False,
                    'model': None,
                    'autocomplete_curtailment': False,
                    'autocompletion': True,
                    'no_output': False,
                    'dev_mode': False,
                    'data_path': None,
                    'sql_connector': None,
                    'sc_inp': None,
                    'cl_out': None,
                    'db': None,
                    }

        defaults.update(kwargs)
        for key, val in defaults.items():
            setattr(self, key, val)
        self.__dict__.update(kwargs)

        self._coldict = aql.get_coldict()

    def read_model_data(self):
        '''
        Read all input data and generate :class:`ModelBase` instance
        attributes.

        '''

        tbrd = TableReader(self.sql_connector, self.sc_inp,
                           self.data_path, self.model)

        # unfiltered input
        dict_tb_2 = {'def_month': [], 'def_week': [],
                     'parameter_month': [], 'tm_soy': []}
        tbrd.df_from_dict(dict_tb_2)

        # read input data filtered by node and energy carrier
        _flt_nd = ([('nd', self.model.slct_node)]
                   if self.model.slct_node else [])
        _flt_ca = ([('ca', self.model.slct_encar)]
                   if self.model.slct_encar else [])
        _flt_pt = ([('pt', self.model.slct_pp_type)]
                   if self.model.slct_pp_type else [])
        dict_tb_3 = {'def_node': _flt_nd,
                     'def_pp_type': _flt_pt,
                     'def_encar': _flt_ca}
        tbrd.df_from_dict(dict_tb_3)

        # translate slct_node_connect to nd_ids
        if self.model.slct_node_connect:
            dict_nd = self.model.df_def_node.set_index('nd').nd_id.to_dict()
            slct_node_connect_id = [(dict_nd[nd1], dict_nd[nd2])
                                    for nd1, nd2 in self.model.slct_node_connect
                                    if nd1 in dict_nd and nd2 in dict_nd]
            _flt_ndcnn = [(('nd_id', 'nd_2_id'), slct_node_connect_id)]
        else:
            _flt_ndcnn = []


        # update filters in case the keyword argument slct_node_id holds more
        # nodes than present in the table
        self.model.slct_node_id = self.model.df_def_node.nd_id.tolist()
        self.model.slct_encar_id = self.model.df_def_encar.ca_id.tolist()
        self.model.slct_pp_type_id = self.model.df_def_pp_type.pt_id.tolist()
        _flt_nd = [('nd_id', self.model.slct_node_id)]
        _flt_ca = [('ca_id', self.model.slct_encar_id)]
        _flt_nd_2 = [('nd_2_id', self.model.df_def_node.nd_id.tolist())]
        _flt_pt = [('pt_id', self.model.df_def_pp_type.pt_id.tolist())]

        # read input data filtered by node, energy carrier, and fuel
        dict_tb_0 = {'def_plant': _flt_nd + _flt_pt,
                     'profchp': _flt_nd,
                     'node_encar': _flt_nd + _flt_ca,
                     'node_connect': _flt_nd + _flt_ca + _flt_nd_2 + _flt_ndcnn}
        tbrd.df_from_dict(dict_tb_0)

        # secondary filtering by plant
        _flt_pp = [('pp_id', self.model.df_def_plant['pp_id'].tolist())]
        _flt_fl = [('fl_id', self.model.df_def_plant.fl_id.unique().tolist())]
        dict_tb_1 = {'profinflow': _flt_pp,
                     'plant_encar': _flt_pp + _flt_ca,
                     'hydro': _flt_pp,
                     'def_fuel': _flt_fl,
                     'plant_month': _flt_pp,
                     'plant_week': _flt_pp,
                     'fuel_node_encar': _flt_fl + _flt_nd + _flt_ca}
        tbrd.df_from_dict(dict_tb_1)

        # initialize profile index dicts
        self.model._init_pf_dicts()

        _flt_pf_supply = [('supply_pf_id',
                          list(self.model.dict_supply_pf.values()))]
        _flt_pf_dmnd = [('dmnd_pf_id',
                        list(self.model.dict_dmnd_pf.values()))]
        _flt_pf_price = [('price_pf_id',
                         list(self.model.dict_pricebuy_pf.values())
                         + list(self.model.dict_pricesll_pf.values()))]
        dict_pf_0 = {'profsupply': _flt_pf_supply,
                     'profdmnd': _flt_pf_dmnd,
                     'profprice': _flt_pf_price,
                     }
        tbrd.df_from_dict(dict_pf_0)

        _flt_pf = [('pf_id', (_flt_pf_price[-1][-1] + _flt_pf_dmnd[-1][-1]
                              + _flt_pf_supply[-1][-1]))]
        dict_pf_1 = {'def_profile': _flt_pf}
        tbrd.df_from_dict(dict_pf_1)

        # filter plants requiring input from non-existing ca
        # e.g. if a fuel-cell is in the input table but no hydrogen is
        # included in the model, the plant's H2 demand wouldn't be accounted
        # for;
        if 'fl_id' in self.model.df_def_encar.columns:

            fl_id_ca = self.model.df_def_encar.fl_id.tolist()
            mask_del = (self.model.df_def_fuel.is_ca.isin([1])
                        & - self.model.df_def_fuel.fl_id.isin(fl_id_ca))

            self.model.df_def_fuel = self.model.df_def_fuel.loc[-mask_del]

        # filter table by special index name/id columns
        self.model.df_parameter_month = \
            self.filter_by_name_id_cols('df_parameter_month',
                                        _flt_fl + _flt_nd + _flt_pp + _flt_ca)

        self._split_profprice()

        # autocomplete input tables
        self.data_autocompletion()

        if isinstance(self.model.df_node_connect, pd.DataFrame):
            self._fix_df_node_connect()

        self.input_table_list = (list(dict_tb_1) + list(dict_tb_2)
                                 + list(dict_tb_0) + list(dict_tb_3)
                                 + list(dict_pf_1))

#        self.model._update_slct_lists()


    def _split_profprice(self):
        '''
        Make two separate DataFrames for profprice buying and selling.

        Having both in the same table gets too complicated down the road.

        '''

        bool_exists = (hasattr(self.model, 'df_profprice')
                       and self.model.df_profprice is not None)

        for bs in ['buy', 'sll']:

            tb_name = 'df_profprice%s' % bs

            if bool_exists:
                mask = self.model.df_def_profile.pf.str.contains('price' + bs)
                list_pd_id = self.model.df_def_profile.loc[mask].pf_id.tolist()
                mask_prf = self.model.df_profprice.price_pf_id.isin(list_pd_id)
                df_split = self.model.df_profprice.loc[mask_prf]
                setattr(self.model, tb_name, df_split)
            else:
                setattr(self.model, tb_name, None)

    def data_autocompletion(self):

        if self.autocompletion:
            logger.info('#' * 60)

            ac.AutoCompletePpType(self.model, self.autocomplete_curtailment)
            ac.AutoCompleteFuelTrns(self.model)
            ac.AutoCompleteFuelDmnd(self.model, self.autocomplete_curtailment)
            ac.AutoCompletePlantTrns(self.model)
            ac.AutoCompletePlantDmnd(self.model, self.autocomplete_curtailment)
            if 'fl_id' in self.model.df_def_encar:
                ac.AutoCompletePlantCons(self.model)
            ac.AutoCompletePpCaFlex(self.model, self.autocomplete_curtailment)
            logger.info('#' * 60)


    def filter_by_name_id_cols(self, name_df, filt):
        """
        Filter a pandas DataFrame with index names in columns.

        This operates on pandas DataFrames where the indices are not provided
        as column names but as row entries in special columns.
        E.g., instead of

        ===== ===== =====
        nd_id fl_id value
        ===== ===== =====
        1     2     1.2
        ===== ===== =====

        we have

        ========== ========== ======== ======== =======
        set_1_name set_2_name set_1_id set_2_id value
        ========== ========== ======== ======== =======
        nd_id      fl_id      1        2        1.2
        ========== ========== ======== ======== =======

        This allows to combine structurally different tables.

        Filtering is implemented as an iteration over the set_n_name/set_n_id
        column pairs, each of which is filtered with respect to all elements
        in the filt parameter.

        Parameters
        ==========

        df : pandas DataFrame
            as described above
        filt : list
            filtering list of the same format as the
            :func:`grimsel.auxiliary.sqlutils.aux_sql_func.read_sql`
            parameter

        Returns
        =======
            filtered DataFrame

        """

        df = getattr(self.model, name_df)

        if df is not None:

            # perform iterative filtering for each name/id column pair
            for name_col, id_col in [(cc, cc.replace('name', 'id'))
                                     for cc in df.columns if 'name' in cc]:
                # init mask
                mask = False

                # loop over all filter elements
                iflt = filt[0]
                for iflt in filt:
                    mask |= (  # 1. select value for current set_n_name column
                             ((df[name_col] == iflt[0])
                              # 2. select corresponding values for set_n_id col
                               & (df[id_col].isin(iflt[1])))
                              # 3. skip if this name_col is not relevant
                             | (df[name_col] == 'na'))

            # reporting
            report_df = df[[c for c in df.columns if 'set_' in c]]
            report_df = report_df.drop_duplicates()

            # get stackable columns
            newcols = [tuple(c.split('_')[1:]) if '_' in c else (None, c)
                       for c in report_df.columns]
            report_df.columns = pd.MultiIndex.from_tuples(newcols,
                                                          names=['col_n',
                                                                 'col_type'])
            report_df = report_df.stack(level='col_n')

            # get all name values from all name columns
            names = df[[c for c in df.columns if 'name' in c]]
            names = names.stack().unique().tolist()
            names = [nn for nn in names if not nn == 'na']

            # get dictionary {set names: set values}
            names_dict = {nn: list(set(report_df.loc[report_df.name == nn, 'id']))
                          for nn in names}

            logger.info('Ex-post filtering of DataFrame {}:'.format(name_df))
            for kk, vv in names_dict.items():
                logger.info('\tSet {} is in ({})'.format(kk, ', '.join(map(str, vv))))

        return df

    def _fix_df_node_connect(self):
        '''
        Makes sure the table df_node_connect corresponds to the new style.

        New style: The transmission capacities are expressed as

        * ``cap_trme_leg`` for exports and
        * ``cap_trmi_leg`` for imports

        for single directions, i.e. non-redundant. The input table has columns
        (nd_id, nd_2_id, ca_id, mt_id, cap_trme_leg, cap_trmi_leg).

        Old style: Single transmission capacity for both directions; columns:
        (nd_id, nd_2_id, ca_id, mt_id, eff, cap_trm_leg)
        '''

        if 'cap_trm_leg' in self.model.df_node_connect.columns:

            df = self.model.df_node_connect

            df['dir'] = df.nd_id < df.nd_2_id

            df_e = df.loc[df.dir].assign(nd_id=df.nd_2_id,
                                         nd_2_id=df.nd_id,
                                         cap_trmi_leg=df.cap_trm_leg)
            df = df.loc[-df.dir].assign(cap_trme_leg=df.cap_trm_leg)
            dfn = pd.concat([df_e, df], sort=False)
            dfn = dfn.drop('cap_trm_leg', axis=1).fillna(0)
            idx = ['nd_id', 'nd_2_id', 'ca_id', 'mt_id']
            logger.info('Aggregation count in fix_df_node_connect:\n',
                        dfn.pivot_table(index=idx, aggfunc=[min, max, len],
                                        values=['cap_trme_leg',
                                                'cap_trmi_leg']))
            dfn = dfn.pivot_table(index=idx, aggfunc=sum,
                                  values=['cap_trme_leg', 'cap_trmi_leg'])

            self.model.df_node_connect = dfn.reset_index()

    @skip_if_resume_loop
    @skip_if_no_output
    def _write_input_tables_to_output_schema(self, tb_list):
        '''
        Gathers relevant input tables and writes them to the output collection.

        Note: As of now profile input tables are excluded. All relevant data
        is included in the parameter output anyway.

        '''

        for itb in set(tb_list) - set(list(zip(*self.runtime_tables))[0]):

            df = getattr(self.model, 'df_' + itb)

            if (df is not None and ('def_' in itb or 'prof' not in itb)):

                log_str = 'Writing input table {} to {} output: {}.'
                logger.info(log_str.format(itb, self.output_target,
                                           self.cl_out))

                if self.output_target == 'psql':

                    logger.info('Writing table {} to output.'.format(itb))
                    engine = self.sql_connector.get_sqlalchemy_engine()
                    db = self.sql_connector.db
                    aql.write_sql(df, db, self.cl_out, itb,
                                  if_exists='replace', engine=engine)

                elif self.output_target == 'hdf5':

                    self.write_hdf(itb, df, 'put')

                elif self.output_target in ['fastparquet']:

                    fn = os.path.join(self.cl_out, itb + '.parq')
                    self.write_parquet(fn, df, self.output_target)
                else:
                    raise RuntimeError('_write_input_tables_to_output_schema: '
                                       'no output_target applicable')

    @skip_if_resume_loop
    @skip_if_no_output
    def write_runtime_tables(self):
        '''
        Some input tables depend on model parameters (time resolution).
        Write these to output database schema.
        Also, table def_node altered due to addition of column dmnd_max.

        '''

        self._write_input_tables_to_output_schema(self.input_table_list)

        skip_fks = [('tm_soy', 'sy'),  # defines sy
                    ('hoy_soy', 'hy')]  # defines hy

        tb_name, pk = ('hoy_soy', ['hy', 'tm_id'])
        for tb_name, pk in self.runtime_tables:

            if getattr(self.model, 'df_' + tb_name, None) is not None:
                df = getattr(self.model, 'df_' + tb_name)

                logger.info('Writing runtime table ' + tb_name)

                cols = []
                for c in df.columns:
                    col_add = [c]

                    if c not in self._coldict:  # same as "value"
                        self._coldict[c] = self._coldict['value']

                    col_add += (list(self._coldict[c])
                                if (tb_name, c) not in skip_fks
                                else list(self._coldict[c][:1]))
                    cols.append(tuple(col_add))

                if self.output_target == 'psql':

                    engine = self.sql_connector.get_sqlalchemy_engine()
                    con_cur = self.sql_connector.get_pg_con_cur()

                    aql.init_table(tb_name=tb_name, cols=cols,
                                   schema=self.cl_out,
                                   ref_schema=self.cl_out, pk=pk, unique=[],
                                   db=self.sql_connector.db, con_cur=con_cur)

                    aql.write_sql(df, sc=self.cl_out, tb=tb_name,
                                  if_exists='append', engine=engine,
                                  con_cur=con_cur)

                elif self.output_target == 'hdf5':

                    self.write_hdf(tb_name, df, 'put')

                elif self.output_target in ['fastparquet']:

                    fn = os.path.join(self.cl_out, tb_name + '.parq')

                    self.write_parquet(fn, df, engine=self.output_target)

                else:
                    raise RuntimeError('write_runtime_tables: no '
                                       'output_target applicable')

    @staticmethod
    def _get_max_run_id(cl_out):

        run_fn = os.path.join(cl_out, 'def_run.parq')

        if os.path.isfile(run_fn):
            logger.debug('_get_max_run_id: Reading file %s.'%run_fn)
            max_run = pd.read_parquet(run_fn, columns=['run_id']).run_id.max()
        else:
            logger.warning('_get_max_run_id: %s not found.'%run_fn)
            max_run = False

        return max_run

class IO:
    '''
    Primary IO class exposing the :module:`io` module.

    :ivar datrd: :class:`DataReader` instance
    :ivar modwr: :class:`ModelWriter` instance
    '''

    def __init__(self, **kwargs):

        self._close_all_hdf_connections()

        defaults = {'sc_warmstart': False,
                    'resume_loop': False,
                    'replace_runs_if_exist': False,
                    'model': None,
                    'autocomplete_curtailment': False,
                    'sql_connector': None,
                    'autocompletion': True,
                    'no_output': False,
                    'dev_mode': False,
                    'data_path': None,
                    'sc_inp': None,
                    'cl_out': None,
                    'db': 'postgres',
                    'output_target': 'psql'
                    }

        defaults.update(kwargs)

        self.sql_connector = defaults['sql_connector']
        self.replace_runs_if_exist = defaults['replace_runs_if_exist']
        self.db = self.sql_connector.db if self.sql_connector else None
        self.cl_out = defaults['cl_out']

        if defaults['resume_loop'] == 'auto':
            defaults['resume_loop'] = \
                    self._get_auto_resume_loop(defaults['output_target'])

        self.resume_loop = defaults['resume_loop']

        self.datrd = DataReader(**defaults)
        self.modwr = ModelWriter(**defaults)

    def _get_auto_resume_loop(self, output_target):

        if not output_target == 'fastparquet':
            raise RuntimeError ('resume_loop="auto" not implemented for '
                                '%s output.')%output_target

        resloop = DataReader._get_max_run_id(self.cl_out)
        resloop = (resloop + 1) if not isinstance(resloop, bool) else resloop

        logger.info('Setting "auto" resume_loop to %s'%resloop)

        return resloop

    @classmethod
    def variab_to_df(cls, py_obj, sets=None):
        ''' Wrapper for backward compatibility. '''

        if not sets:
            sets = table_struct.DICT_COMP_IDX[py_obj.name]
            sets = [st for st in sets if not st == 'bool_out']

        return VariabIO._to_df(py_obj, sets)

    @classmethod
    def param_to_df(cls, py_obj, sets=None):
        ''' Wrapper for backward compatibility. '''

        if not sets:
            sets = table_struct.DICT_COMP_IDX[py_obj.name]

        return ParamIO._to_df(py_obj, sets)

    def read_model_data(self):

        self.datrd.read_model_data()

    def write_runtime_tables(self):

        self.datrd.write_runtime_tables()

    def init_output_tables(self):

        self.modwr.init_compio_objs()
        self.modwr.init_all()

    def write_run(self, run_id):

        self.modwr.run_id = run_id
        self.modwr.write_all()

    def _init_loop_table(self, cols_id, cols_step, cols_val):

        tb_name = 'def_run'
        cols = ([('tdiff_solve', 'DOUBLE PRECISION'),
                 ('tdiff_write', 'DOUBLE PRECISION'),
                 ('run_id', 'SMALLINT'),
                 ]
                + [(s, 'SMALLINT') for s in cols_id]
                + [(s, 'DOUBLE PRECISION') for s in cols_step]
                + [(s, 'VARCHAR(30)') for s in cols_val]
                + [('info', 'VARCHAR'), ('objective', 'DOUBLE PRECISION')])

        if self.modwr.output_target == 'psql':

            aql.init_table(tb_name, cols, self.cl_out,
                           pk=cols_id, unique=['run_id'], db=self.db)

        elif self.modwr.output_target == 'hdf5':

            df = pd.DataFrame(columns=list(zip(*cols))[0])

            if self.modwr.output_target == 'hdf5':
                df.to_hdf(self.cl_out, tb_name, format='table')

        elif self.modwr.output_target == 'fastparquet':
            pass  # parquet table is not initialized

        else:
            raise RuntimeError('_init_loop_table: no '
                               'output_target applicable')

    @staticmethod
    def _close_all_hdf_connections():

        tables.file._open_files.close_all()


# %%
if __name__ == '__main__':

    pass



'''
Auxiliary functions for database manipulation.
TODO: Make this a class with a single db connection.
'''

import psycopg2 as pg
import time
import pandas as pd
from sqlalchemy import create_engine
import sqlalchemy
from collections import OrderedDict


import subprocess
from os import walk
try:
    # Linux only
    from sh import pg_dump
except:
    pass
import os

import grimsel.grimsel_config as config

# %%



def close_con(conn):
    conn.close()


class SqlConnector():
    '''
    Manages connection strings for psycopg2 and sqlalchemy, open and closes
    connections.

    TODO: Could pass config module to __init__ for config_dict initialization.

    '''

    def __init__(self, db, **kwargs):

        self.db = db

        config_dict = {
        'user': config.PSQL_USER,
        'password': config.PSQL_PASSWORD,
        'host': config.PSQL_HOST,
        'port': config.PSQL_PORT,
        }
        for kw, val in config_dict.items():
            setattr(self, kw, val)
        self.__dict__.update(**kwargs)

        self.pg_str = ('dbname={db} user={user} password={password} '
                      'host={host}').format(**self.__dict__)

        self.sqlal_str = ('postgresql://{user}:{password}'
                          '@{host}:{port}/{db}').format(**self.__dict__)

        self._sqlalchemy_engine = None
        self._conn = None
        self._cur = None

    def get_sqlalchemy_engine(self):

        if not self._sqlalchemy_engine:
            self._sqlalchemy_engine = create_engine(self.sqlal_str)

        return self._sqlalchemy_engine

    def get_pg_con_cur(self):

        if not self._conn:
            self._conn = pg.connect(self.pg_str)
            self._cur = self._conn.cursor()

        return self._conn, self._cur

    def __repr__(self):
        strg = '%s\n%s'%(self.pg_str, self.sqlal_str)
        return(strg)


def exec_sql(exec_str, ret_res=True, time_msg=False, db=None, con_cur=None):
    t = time.time()
    ''' Pass sql query to the server. '''

    conn, cur = (SqlConnector(db).get_pg_con_cur()
                 if not con_cur else con_cur)

    cur.execute(exec_str)
    conn.commit()
    result = None
    try:
        result = cur.fetchall()
    except:
        pass
#        print("No results")

    if con_cur is None:
        conn.close()

    if time_msg:
        print(time_msg, time.time() - t, 'seconds')

    if ret_res:
        return result


# %%

def get_column_string(cols, kind):
    '''
    Generates strings for the multiple manipulation of columns.

    Parameters:
    cols -- columns to be included in the string;
            for kind in (drop, set): list; for kind in (add): dictionary
            of shape {name: datatype}
    kind -- string, one of ('drop', 'add', 'set')
    '''

    if kind == 'drop':
        return ', '.join(['DROP COLUMN IF EXISTS {}'.format(col)
                         for col in cols])
    elif kind == 'add':
        return ', '.join(['ADD COLUMN {} {}'.format(col, dt)
                         for col, dt in cols.items()])
    elif kind == 'set':
        return 'SET ' + ','.join(['{col} = temp.{col}'.format(col=col)
                                  for col in cols])


# %%

def get_sql_tables(sc, db=None, con_cur=None):
    '''
    Returns list of tables in schema
    '''

    exec_str = '''
               SELECT table_name FROM
               information_schema.tables
               WHERE table_schema = \'{sc}\'
               AND table_type = \'BASE TABLE\'
               '''.format(sc=sc)
    return [itb[0] for itb in exec_sql(exec_str, db=db, con_cur=con_cur)]


# %%

def get_sql_cols(tb, sc='public', db=None, con_cur=None):
    '''
    Returns the names and data types of the selected table as a dictionary
    {'col_name': 'col_type', ...}
    '''

    exec_str = '''
                SELECT column_name, data_type
                FROM information_schema.columns
                WHERE table_schema = \'{sc}\'
                AND table_name = \'{tb}\'
                '''.format(sc=sc, tb=tb)

    exec_sql_kwargs = {'exec_str': exec_str}
    exec_sql_kwargs.update({'con_cur': con_cur} if con_cur else {'db': db})

    dict_col = OrderedDict(exec_sql(**exec_sql_kwargs))
    return dict_col


# %%

def reset_schema(sc, db, warn=True):

    sc_exist = len(exec_sql('''
                            SELECT schema_name
                            FROM information_schema.schemata
                            WHERE schema_name = \'{sc}\';
                            '''.format(sc=sc), db=db)) > 0

    if sc_exist:
        try:
            max_run_id = exec_sql('''
                                  SELECT MAX(run_id)
                                  FROM {sc}.def_loop
                                  '''.format(sc=sc), db=db)[0][0]
        except:
            # no def_loop
            max_run_id = 'None'


        if warn:
            input(
'''
~~~~~~~~~~~~~~~   WARNING:  ~~~~~~~~~~~~~~~~
You are about to delete existing schema {sc}.
The maximum run_id is {max_run_id}.

Hit enter to proceed.
'''.format(sc=sc, max_run_id=max_run_id)
)


    print('Dropping output schema: ', sc)
    # This could be a call to a dedicated function init_schema
    # in a dedicated module (like aux_sql_func)
    exec_sql('''
             DROP SCHEMA IF EXISTS {sc_out} CASCADE;
             CREATE SCHEMA IF NOT EXISTS {sc_out}
             '''.format(sc_out=sc), db=db)



# %%

def write_sql(df, db=None, sc=None, tb=None, if_exists=None,
              engine=None, chunksize=None, con_cur=None):


    if not engine:
        sqlc = SqlConnector(db)
        _engine = sqlc.get_sqlalchemy_engine()
    else:
        _engine = engine

    if con_cur:
        exec_sql_kwargs = dict(con_cur=con_cur)
    else:
        exec_sql_kwargs = dict(db=db)



    if if_exists == 'replace':
        exec_str = ('''DROP TABLE IF EXISTS {sc}.{tb} CASCADE;
                    ''').format(sc=sc, tb=tb)
        exec_sql(**dict(exec_str=exec_str, **exec_sql_kwargs))
    else:
        # add columns which exist in the source table but not in
        # the database table;
        # using some internal pandas methods to map the pandas
        # datatypes to SQL datatypes

        pandas_sqltable = pd.io.sql.SQLTable('_', _engine, df, schema='_',
                                             index=False)

        dtype_mapper = pandas_sqltable._sqlalchemy_type

        new_cols = pandas_sqltable._get_column_names_and_types(dtype_mapper)

        old_cols = get_sql_cols(tb, sc,
                                **({'con_cur': con_cur}
                                   if con_cur else {'db': db})).keys()

        VisitableType = sqlalchemy.sql.visitors.VisitableType

        new_cols = {'"{}"'.format(c[0]):
                    c[1]().compile()
                        if isinstance(c[1], VisitableType)
                        else c[1].compile()
                    for c in new_cols if not c[0] in old_cols}
        add_str = get_column_string(new_cols, 'add')

        if not add_str == '':
            exec_strg = '''
                        CREATE TABLE IF NOT EXISTS {sc}.{tb}();
                        ALTER TABLE {sc}.{tb}
                        {add_str};
                        '''.format(sc=sc, tb=tb, add_str=add_str)
            exec_sql(**dict(exec_str=exec_strg, **exec_sql_kwargs))

    df.to_sql(name=tb, con=_engine, schema=sc, if_exists=if_exists,
              index=False, chunksize=chunksize)

    if not engine:
        _engine.dispose()

# %%


def write_sql_grouped(df, db, sc, tb, if_exists, by, verbose=True):
    '''
    Writes DataFrames by slices to avoid the overload due to conversion
    of the DataFrame to an SQL string.

    Parameters:
    df -- DataFrame to be written
    db -- Target database
    sc -- Target schema
    tb -- Target schema
    if_exists -- string ('append', 'replace')
    by -- the by parameter of pandas.DataFrame.groupby
    '''

    dfgrpd = df.groupby(by)

    for ngroup, group in enumerate(dfgrpd.groups):
        if verbose:
            print(('Writing group {} to table {}, number {} of {}')
                  .format(group, sc + '.' + tb, ngroup + 1,
                          len(dfgrpd.groups)))

        dfg = dfgrpd.get_group(group)

        _if_exists = if_exists if ngroup is 0 else 'append'

        write_sql(dfg, db, sc, tb, if_exists=_if_exists)





# %%
def assemble_filt_sql(filt, func):
    '''
    Takes an appropriately shaped filt list and returns
    the argument for the SQL call to WHERE.
    args:
    filt -- list like [('columns', ['value1', 'value2'], binary_operator,
                                 secondary_binary_operator), (next col, ...)]
    func -- dict like {'columns': 'ABS'}
    Example:
    filt=[('run_id', [1, 3]),
          ('pp', ['%WIN%', '%SOL%', '%HCO%'], ' NOT LIKE ', ' AND '),
          ('nd', ['DE0', 'AT0'], ' <> ')]
    returns:
        (nd <> 'DE0' OR nd <> 'AT0')
            AND (run_id = 1 OR run_id = 3)
            AND (pp NOT LIKE '%WIN%' AND pp NOT LIKE '%SOL%'
                 AND pp NOT LIKE '%HCO%')
    Note how ('nd', ['DE0', 'AT0'], ' <> ') has no effect due to the
    default 'OR', hence the necessity for the explicit
    secondary_binary_operator.
    Note: empty value sublists mean all (filter ignored).
    '''

    filt = [list(f) if len(f) >= 3 else list(f) + [' = '] for f in filt]
    filt = [list(f) if len(f) >= 4 else list(f) + [' OR '] for f in filt]
    filt = [f for f in filt if not len(f[1]) is 0]

    rel = {f[0]: f[2] for f in filt}
    opr = {f[0]: f[3] for f in filt}
    fnc = {f[0]: func[f[0]] if f[0] in func.keys() else '' for f in filt}

    filt_dict = {f[0]: [('\'' + if1 + '\'') for if1 in f[1]]
                 if type(f[1][0]) == str else f[1] for f in filt}
    filt_str = ' AND '.join(['(' + opr[ky].join([fnc[ky] + '(' + ky + ')' + rel[ky] + str(ivl)
                                                for ivl in vl]) + ')'
                             for ky, vl in filt_dict.items()])
    return filt_str
#
#if __name__ == '__main__':
#    filt=[('run_id', [1, 3]),
#          ('pp', ['%WIN%', '%SOL%', '%HCO%'], ' NOT LIKE ', ' AND '),
#          ('nd', ['DE0', 'AT0'], ' <> '),
#          ('erg_yr_sy', [1e-6], '<')]
#
#    print(assemble_filt_sql(filt, func={'erg_yr_sy': 'ABS'}))

# %%

def copy_table_structure(sc, tb, sc0, db, verbose=False):
    '''
    Performs the CREATE TABLE LIKE operation.

    Arguments:
    sc -- schema new table
    sc0 -- schema existing table
    tb -- table name
    '''

    exec_str = ('''
                DROP TABLE IF EXISTS {sc}.{tb} CASCADE;
                CREATE TABLE {sc}.{tb} (LIKE {sc0}.{tb} INCLUDING ALL);
                ''').format(sc=sc, tb=tb, sc0=sc0)
    if verbose:
        print(exec_str)
    exec_sql(exec_str, db=db)



# %%

# OBSOLETE? read_sql with intermediate temporary table

#def read_sql(db, sc, tb, filt=False, filt_func=False, drop=False, keep=False,
#             copy=False, tweezer=False, distinct=False, verbose=False):
#    '''
#    Keyword arguments:
#    tweezer -- list; filter (exclude or include) single combinations of values
#                tweezer = [' AND ', ({'nd': 'IT0', 'bool_out': True}, ' NOT '),
#                                    ({'nd': 'FR0', 'bool_out': False}, ' NOT ')]
#    '''
#
#    sqlc = sql_connector(db)
#    engine = sqlc.get_sqlalchemy_engine()
#
#    if verbose:
#        print('Reading ' + sc + '.' + tb +  ' from database ' + db)
#        print('filt=', filt)
#    tweez_str = ''
#    if tweezer:
#        if type(tweezer[0]) is tuple:
#            tweezer.insert(0, ' AND ')
#
#        tweez_str = ' %s (' %tweezer[0]
#        tweez_str += ' AND '.join([tw[1] + '(' + ' AND '.join([str(kk) + ' = ' + '\'' + str(vv) + '\'' for kk, vv in tw[0].items()]) + ')' for tw in tweezer[1:]])
#        tweez_str += ')'
#        if verbose:
#            print('tweez_str', tweez_str)
#
#    # filtering through sql
#    if filt:
#
#        temp_tb = ('temp_read_sql_filtered_'
#                   + ''.join(random.choice(string.ascii_lowercase)
#                             for _ in range(10)))
#        if not filt_func:
#            filt_func = {}
#
#        distinct_str = 'DISTINCT ' if distinct else ''
#        keep_str = ', '.join(keep) if keep else '*'
#        filt_str = assemble_filt_sql(filt, filt_func)
#        exec_str = ('''
#                    DROP TABLE IF EXISTS {temp_tb} CASCADE;
#                    SELECT {distinct_str} {keep_str}
#                    INTO {temp_tb}
#                    FROM {sc}.{tb}
#                    WHERE {filt_str}
#                          {tweez_str};
#                    ''').format(keep_str=keep_str, sc=sc, tb=tb,
#                                filt_str=filt_str, tweez_str=tweez_str,
#                                distinct_str=distinct_str, temp_tb=temp_tb)
#        if verbose:
#            print(exec_str)
#        exec_sql(exec_str, db=db)
#        df = pd.read_sql_table(temp_tb, engine, schema='public')
#
#        exec_sql('DROP TABLE {};'.format(temp_tb), db=db)
#
#    else:
#        df = pd.read_sql_table(tb, engine, schema=sc)
#        if keep:
#            df = df.loc[:, keep]
#
#        if distinct:
#            df = df.drop_duplicates()
#
#    if copy:
#        copy_table_structure(sc=copy, tb=tb, sc0=sc, db=db, verbose=verbose)
##        exec_str = ('''
##                    DROP TABLE IF EXISTS {sc}.{tb} CASCADE;
##                    CREATE TABLE {sc}.{tb} (LIKE {sc0}.{tb} INCLUDING ALL);
##                    ''').format(sc=copy, tb=tb, sc0=sc)
##        if verbose:
##            print(exec_str)
##        exec_sql(exec_str, db=db)
#        write_sql(df, db, copy, tb, 'append')
#
#    if drop:
#        df = df.drop(drop, axis=1)
#
#    engine.dispose()
#
#    return df



def read_sql(db=None, sc=None, tb=None, filt=False, filt_func=False, drop=False,
             keep=False, copy=False, tweezer=False, distinct=False, verbose=False,
             engine=None, limit=None):
    '''
    Keyword arguments:
    tweezer -- list; filter (exclude or include) single combinations of values
                tweezer = [' AND ', ({'nd': 'IT0', 'bool_out': True}, ' NOT '),
                                    ({'nd': 'FR0', 'bool_out': False}, ' NOT ')]
    '''

    if not engine:
        sqlc = SqlConnector(db)
        _engine = sqlc.get_sqlalchemy_engine()
    else:
        _engine = engine


    limit_str = 'LIMIT %d'%limit if isinstance(limit, int) else ''

    if verbose:
        print('Reading ' + sc + '.' + tb +  ' from database ' + db)
        print('filt=', filt)
    tweez_str = ''
    if tweezer:
        if type(tweezer[0]) is tuple:
            tweezer.insert(0, ' AND ')

        tweez_str = ' %s (' %tweezer[0]
        tweez_str += ' AND '.join([tw[1] + '(' + ' AND '.join([str(kk) + ' = ' + '\'' + str(vv) + '\'' for kk, vv in tw[0].items()]) + ')' for tw in tweezer[1:]])
        tweez_str += ')'
        if verbose:
            print('tweez_str', tweez_str)

    cols = keep if keep else get_sql_cols(tb, sc, db)

    # filtering through sql
    if filt and not any(not list(ff[1]) for ff in filt):

        if not filt_func:
            filt_func = {}

        distinct_str = 'DISTINCT ' if distinct else ''
        keep_str = ', '.join(keep) if keep else '*'
        filt_str = assemble_filt_sql(filt, filt_func)
        exec_str = ('''
                    SELECT {distinct_str} {keep_str}
                    FROM {sc}.{tb}
                    WHERE {filt_str}
                          {tweez_str}
                    {limit_str};
                    ''').format(keep_str=keep_str, sc=sc, tb=tb,
                                filt_str=filt_str, tweez_str=tweez_str,
                                distinct_str=distinct_str,
                                limit_str=limit_str)
        if verbose:
            print(exec_str)

        df = pd.DataFrame(exec_sql(exec_str, db=db), columns=cols)

    elif filt:
        # one of the value lists is empty... return empty DataFrame
        df = pd.DataFrame(columns=cols)

    else:
        df = pd.read_sql_table(tb, _engine, schema=sc)
        if keep:
            df = df.loc[:, keep]

        if distinct:
            df = df.drop_duplicates()

    if copy:
        copy_table_structure(sc=copy, tb=tb, sc0=sc, db=db, verbose=verbose)
        write_sql(df, db, copy, tb, 'append')

    if drop:
        df = df.drop(drop, axis=1)

    if not engine:
        _engine.dispose()

    return df



# %%

def on_str(lst, pl='', pr=''):
    return ' AND '.join(['(' + pl + '.' + c + ' = '
                             + pr + '.'  + c + ')'
                         for c in lst])
# %%


def slct_str(lst, exclude=[], pl='', pr='', sl='', sr='', commas=['','']):
    if lst == []:
        if commas == [',',',']:
            commas = [',','']
        else:
            commas = ['','']
    lst = [l for l in lst if not l in exclude]

    lst_1 = []
    for ilst in lst:

        if type(ilst) == list: # defined AS
            lst_1_add = ((pl if pl else '') + ilst[0] + (sl if sl else ''),
                         (pr if pr else '') + ilst[1] + (sr if sr else ''))
            lst_1.append(' AS '.join(lst_1_add))
        elif type(ilst) == str: # only left
            lst_1.append((pl if pl else '') + ilst + (sl if sl else ''))

    lst_1 = ', '.join(lst_1)


    return commas[0] + lst_1 + commas[1]


# %%

# define dictionary containing SQL data type and foreign key for
# model sets
coldict = {
           # FORMAT:
           # 'set_name': ('DATATYPE', )
           'sy': ('INTEGER', '{sc}.tm_soy(sy)'),
           'hy': ('FLOAT', '{sc}.hoy_soy(hy)'),
           'pp_id': ('SMALLINT', '{sc}.def_plant(pp_id)'),
           'pp': ('VARCHAR', '{sc}.def_plant(pp)'),
           'pt_id': ('SMALLINT', '{sc}.def_pp_type(pt_id)'),
           'pt': ('VARCHAR', '{sc}.def_pp_type(pt)'),
           'ca_id': ('SMALLINT', '{sc}.def_encar(ca_id)'),
           'wk_id': ('SMALLINT', '{sc}.def_week(wk_id)'),
           'value': ('DOUBLE PRECISION',),
           'mt_id': ('SMALLINT', '{sc}.def_month(mt_id)'),
           'month': ('VARCHAR',),
           'fl_id': ('SMALLINT', '{sc}.def_fuel(fl_id)'),
           'fl': ('VARCHAR', '{sc}.def_fuel(fl)'),
           'nd_id': ('SMALLINT', '{sc}.def_node(nd_id)'),
           'nd': ('VARCHAR', '{sc}.def_node(nd)'),
           'nd_2_id': ('SMALLINT', '{sc}.def_node(nd_id)'),
           'bool_out': ('BOOLEAN',),
           'weight': ('FLOAT',),
           'min_hoy': ('SMALLINT',), 'year': ('SMALLINT',),
           'day': ('SMALLINT',),
           'dow': ('SMALLINT',), 'dow_name': ('VARCHAR',),
           'doy': ('SMALLINT',), 'hom': ('SMALLINT',),
           'hour': ('SMALLINT',), 'how': ('SMALLINT',),
           'wk': ('SMALLINT',), 'wk_weight': ('SMALLINT',),
           'dow_type': ('VARCHAR',),
           '"DateTime"': ('TIMESTAMP',),
           'mt': ('VARCHAR',),
           'ndays': ('SMALLINT',),
           'season': ('VARCHAR',),
           'wom': ('SMALLINT',),
           'pwrerg_cat': ('VARCHAR',),
           'tm_id': ('SMALLINT',),
           'run_id': ('SMALLINT', '{sc}.def_loop(run_id)'),
#           'supply_pf_id': ('SMALLINT', '{sc}.def_profile(pf_id)'),
#           'dmnd_pf_id': ('SMALLINT', '{sc}.def_profile(pf_id)'),
#           'pricesll_pf_id': ('SMALLINT', '{sc}.def_profile(pf_id)'),
#           'pricebuy_pf_id': ('SMALLINT', '{sc}.def_profile(pf_id)'),
           'pf_id': ('SMALLINT', '{sc}.def_profile(pf_id)'),
          }

def get_coldict(sc=None, db=None, fk_include_missing=False, con_cur=None):
    ''' Adding SQL schema name to foreign keys in coldict. '''


    if sc:
        _coldict = {}
        for kk, vv in coldict.items():
            _val = tuple(ivv.format(sc=sc) for ivv in vv)
            if len(_val) > 1:
                if (not _val[1].split('.')[1].split('(')[0]
                         in get_sql_tables(sc, db, con_cur=con_cur)):
                    _val = (_val[0],)
            _coldict[kk] = [iv.format(sc) for iv in _val]
    else:
        _coldict = {col: [coltype[0]] for col, coltype in coldict.items()}


    return _coldict


def init_table(tb_name, cols, schema='public', ref_schema=None,
               pk=[], unique=[], appendix='', bool_auto_fk=False,
               bool_return=False, db=None, con_cur=None,
               skip_if_exists=False, warn_if_exists=False):
    '''
    Drop and re-initialize indexed table.
    Keyword arguments:
    - tb_name -- table name
    - cols -- list of table columns plus definition of foreign keys
    - schema -- database schema
    - pk -- list of primary key column names
    - unique -- list of unique column names
    - appendix --
    - bool_auto_fk -- complete foreign keys based on the coldict
    - bool_return -- function returns sql string if True
    '''

    if ref_schema is None:
        ref_schema = schema


    _skip = (True if (skip_if_exists
                      and tb_name in get_sql_tables('log', 'fnnc'))
                 else False)


    def warn_exists():

        if warn_if_exists:

            tb_exists = len(exec_sql('''
                            SELECT table_name
                            FROM information_schema.tables
                            WHERE table_schema = \'{sc}\'
                            AND table_name = \'{tb}\';
                            '''.format(sc=schema, tb=tb_name), db=db)) > 0

            if tb_exists:
                exec_count = 'SELECT COUNT(*) FROM %s.%s'%(schema, tb_name)
                len_tb = exec_sql(exec_count, db=db)[0][0]
                input(
'''
~~~~~~~~~~~~~~~   WARNING:  ~~~~~~~~~~~~~~~~
You are about to delete existing table %s.%s.
The length is %d.

Hit enter to proceed.
'''%(schema, tb_name, len_tb))




    if _skip:
        print('Table {} in schema {} exists: skipping init.'.format(tb_name,
                                                                    schema))
    else:

        warn_exists()

        # apply default data type and foreign key to the columns if not provided
        _coldict = get_coldict(ref_schema, db, con_cur=con_cur)

        if not bool_auto_fk:
            _coldict = {col: (typeref[0],)
                        for col, typeref in _coldict.items()}

        _cols = []
        for icol in cols:
            if not (type(icol) is tuple or type(icol) is list):
                icol = (icol,)

            if len(icol) is 1: # column must be in _coldict, otherwise ill-defined
                icol_new = (*icol, *_coldict[icol[0]])
            elif len(icol) is 2:
                icol_new = icol
            else:
                icol_new = icol # all there
            _cols.append(icol_new)



        exec_str = 'DROP TABLE IF EXISTS ' + schema + '.' + tb_name + ' CASCADE;\n'
        exec_str += ('CREATE TABLE ' + schema + '.' + tb_name + ' (')
        exec_str += ',\n '.join([icol[0] + ' ' + icol[1] +
                               (' UNIQUE' if icol[0] in unique else '') +
                               (' REFERENCES ' + icol[2] if len(icol) > 2 else '')
                               for icol in _cols])
        exec_str += ', PRIMARY KEY (' + ', '.join(pk) + ')' if len(pk) > 0 else ''
        exec_str += ', ' + appendix if appendix != '' else ''
        exec_str += ')'

        exec_sql(exec_str, db=db, con_cur=con_cur)

        return ', '.join([c[0] for c in cols])




# %%

def filtered_aggregates(nbin=20,
                        value_table=['out_st_lp', 'duals_supply'],
                        mask_table=['out_st_lp', 'variabs_soy_mod'],
                        out_table=['out_st_lp_analysis', 'filtered_mc_respto_pp'],
                        value_slct={'encar': ['EL'],
                                    'constr': ['supply']},
                        mask_slct={'encar': ['EL'],
                                   'variable': ['power'],
                                   'fuel': ['pumped_hydro_charg',
                                                'pumped_hydro']},
                        merge_cols=['soy', 'node', 'year'],
                        weight_type='filtered_histogram'
                        ):

    nbin=20
    value_table=[out_sc, 'duals_supply']
    mask_table=[out_sc, 'var_sy_pwr']
    out_table=['public', 'filtered_mc_respto_pp']
    value_slct={}
    mask_slct={'pp_type': ['CAS_STO',
                           'LIO_STO']}
    merge_cols=['soy', 'node']
    weight_type='filtered_histogram'

    """
    Obtain filtered histograms or filtered/weighted averages from PSQL DB

    Keyword arguments:
    nbin -- number of bins between min and max of the respective data.
    value_table -- ['schema', 'table_name']
    mask_table -- ['schema', 'table_name']
    out_table -- ['schema', 'table_name']
    value_slct -- dictionary {'column': ['value1', 'value2']} for value table
    mask_slct -- dictionary {'column': ['value1', 'value2']} for mask table
    merge_cols -- list of columns to merge the two tables on
    weight_type -- ['filtered_histogram', ...]

    """

#
    t = time.time()

    # derived
    val_cat = [c for c in value_slct]# if not len(value_slct[c]) == 1]
    cols_slct_val = (['value']
                     + val_cat
                     + merge_cols)
    mask_cat = [c for c in mask_slct]# if not len(mask_slct[c]) == 1]
    cols_slct_mask = (['value']
                      + mask_cat
                      + merge_cols)

    def slct_str(lst, exclude=[], prefix=False, suffix_as=False, suffix=False,
                 commas=['','']):
        if lst == []:
            if commas == [',',',']:
                commas = [',','']
            else:
                commas = ['','']
        main = ', '.join([(prefix if prefix else '')
                           + l + (suffix if suffix else '')
                           + (' AS ' + l + suffix_as if suffix_as else '')
                           for l in lst if not l in exclude])
        return commas[0] + main + commas[1]

    def dict_to_where(dct):
        '''
        Takes input dictionary {column: [list of values]} and transforms it into
        an SQL WHERE statement
        '''
        return (' AND '.join(['(' + ' OR '.join([ivr + ' = ' + '\'' + ivl + '\''
                                                 for ivl in dct[ivr]]) + ')'
                              for ivr in dct]))
    def on_str(lst):
        return ' AND '.join(['tm.' + c + '_mask = ' + 'tv.' + c + '_val' for c in lst])


    exec_str = (
    'DROP TABLE IF EXISTS temp_mrg; SELECT '
    + ', '.join(['tv.' + c + '_val AS ' + c for c in cols_slct_val]) + ', '
    + slct_str([c + '_mask' for c in cols_slct_mask if not c in  merge_cols],
                prefix='tm.') +
    ' INTO temp_mrg' +
    ' FROM ( SELECT ' + slct_str(cols_slct_val, suffix_as='_val') +
    ' FROM ' + '.'.join(value_table) +
    ' WHERE ' + dict_to_where(value_slct) + ' ) tv ' +
    ' LEFT JOIN ( SELECT ' +
    slct_str(cols_slct_mask, suffix_as='_mask', prefix='tm.') +
    ' FROM ( SELECT ' + slct_str(cols_slct_mask) + ' FROM ' +
    '.'.join(mask_table) +
    ' WHERE ' + dict_to_where(mask_slct) +
    ') tm ' +
    ') tm ' +
    ' ON ' + on_str(merge_cols) +
    ' WHERE NOT value_mask IS NULL; ' +
    ' DROP TABLE IF EXISTS ' + '.'.join(out_table) + ';'
    )

    #
    #if weight_type == 'bin':
    #    exec_str += (
    #    '''
    #    /* BINARY MASK */
    #    SELECT AVG(value)
    #    '''
    #     + ('' if len(merge_cols) == 0 else ', ') + ', '.join(merge_cols) +
    #    '''
    #    INTO
    #    '''
    #    + '.'.join(out_table) +
    #    '''
    #    FROM temp_mrg
    #    	WHERE mask <> 0
    #        GROUP BY
    #    '''
    #    + ', '.join(merge_cols) + ';'
    #    )
    #
    #
    #
    #if weight_type == 'weighted':
    #    exec_str += (
    #    '''
    #    /* WEIGHTED AVERAGE */
    #    SELECT SUM(mask * value) / SUM(mask) AS weighted_sum
    #    '''
    #    + ('' if len(merge_cols) == 0 else ', ') + ', '.join(merge_cols) +
    #    '''
    #    INTO
    #    '''
    #    + '.'.join(out_table) +
    #    '''
    #    FROM temp_mrg
    #        GROUP BY
    #    '''
    #    + ', '.join(merge_cols) + ';'
    #    )


# %%

#def multijoin(tables, col_select, merge_cols, keep_cols, output_where):
#
#    '''
#    dict tables --- {0: ['schema', 'name', 'alias'], 1: ...,
#                     'out': ['schema', 'name']}
#    dict col_select = {0: ['column_1', ['column_2', 'AS alias']], 1: ...,
#                       'out': ['col_from_input']}
#    dict merge_cols = {0: [], # not relevant by definition
#                       1: ['col_1'], ...}
#    dict keep_cols = {0: ['variable', 'soy', 'node', 'node_2', 'encar',
#                     'year','power_transmission'],
#                     1: ['mc_node',],
#                     2: ['mc_node_2']}
#    list output_where = ['power_transmission <> 0']
#    '''
#
#    ntb = len(tables) - 1
#    exec_str = 'DROP TABLE IF EXISTS ' + '.'.join(tables['out']) + ';'
#    exec_str += ' WITH '
#    for itb in range(ntb):
#        exec_str += (tables[itb][2] +
#                     ' AS ( SELECT ' + slct_str(col_select[itb]) +
#                     ' FROM ' + '.'.join(tables[itb][:2]) + '), '
#                     )
#    exec_str += 'temp_mrg AS (SELECT '
#    for itb in range(0, ntb):
#        exec_str += slct_str(keep_cols[itb], pl=tables[itb][2] + '.')
#        exec_str += ', ' if (itb < ntb - 1) else ' '
#    exec_str += 'FROM ' + tables[0][2] + ' '
#    for itb in range(1, ntb):
#        exec_str += 'LEFT JOIN ' + tables[itb][2] + ' ON '
#
#        exec_str += on_str(merge_cols[itb], pl=tables[0][2],
#                                            pr=tables[itb][2])
#    exec_str += ') SELECT ' + slct_str(col_select['out'])
#    exec_str += ' INTO ' + '.'.join(tables['out'])
#    exec_str += ' FROM temp_mrg '
#    exec_str += (' WHERE ' + ' AND '.join(['(' + o + ')'
#                                           for o in output_where])
#                  if output_where != [] else '')
#
#    conn = pg.connect(get_config('sql_connect')['psycopg2']
#                      .format(db=get_config('sql_connect')['db']))
#    cur = conn.cursor()
#    cur.execute(exec_str)
#    conn.commit()
#    try:
#        cur.fetchall()
#    except:
#        print("No results")
#    conn.close()
#

# %%



def add_column(df_src, tb_tgt, col_new, on_cols, db,
               data_type='DOUBLE PRECISION'):
    '''
    Add column from dataframe to existing table in database.
    Keyword arguments:
    df_src -- dataframe containing the column to be added
    tb_tgt -- target table as list [schema, table]
    col_new -- name of column to be added
    on_cols -- list of column names for table alignment
    '''
#new_cols
#df_src = dfstep
#tb_tgt = ['out_disagg', 'step_evts_all']
#col_new = new_cols
#on_cols
#db = 'storage2'
#data_type = 'VARCHAR'

    if type(col_new) is dict:
        data_type = col_new.copy()
        col_new = list(col_new.keys())
    else:
        if not type(col_new) is list:
            col_new = [col_new]
        if not type(data_type) is dict:
            data_type = {col: data_type for col in col_new}

    # get data into database, then copy within sql
    write_sql(df_src[on_cols + col_new], db, 'public',
              'temp_add_column', 'replace', chunksize=10000)

    where_str = ' AND '.join(['tgt.' + c + ' = temp.' + c for c in on_cols])

    drop_str = get_column_string(col_new, 'drop')
    add_str = get_column_string(data_type, 'add')
    set_str = get_column_string(col_new, 'set')

    exec_str = ('''
                ALTER TABLE {tb_tgt}
                {drop_str};

                ALTER TABLE {tb_tgt}
                {add_str};

                UPDATE {tb_tgt} AS tgt
                {set_str}
                FROM temp_add_column AS temp
                WHERE {where};
                ''').format(drop_str=drop_str, add_str=add_str,
                            set_str=set_str,
                            tb_tgt='.'.join(tb_tgt),
                            col_new=col_new, where=where_str,
                            data_type=data_type)
    print(exec_str)
    exec_sql(exec_str, db=db)

# %%
def joinon(db, new_col, on_col, tb_target, tb_source,
           new_columns=True, verbose=False):
    '''
    Join columns from one PostgreSQL table to another table.

    Parameters:
    db         : database name
    new_col    : list or dict, if dict, then
                 {col1_name_in_source: new_col1_name_target, ...}
    on_col     : list or dict, if dict, then
                 {col1_name_in_source: col1_name_in_target, ...}
    tb_target  : list or str, specifying target table
                 as ['schema', 'table'] or 'schema.table'
    tb_source  : list or str, specifying source table
                 as ['schema', 'table'] or 'schema.table'
    new_columns: boolean; True: generates new columns
    '''

    if not type(new_col) is dict:
        new_col = {c: c for c in new_col}
    if not type(on_col) is dict:
        on_col = {c: c for c in on_col}

    if type(tb_target) is str:
        tb_target = tb_target.split('.')
    if type(tb_source) is str:
        tb_source = tb_source.split('.')

    exec_str = '''
               SELECT column_name, data_type
               FROM information_schema.columns
               WHERE table_name = \'{tb}\'
               AND table_schema = \'{sc}\';
               '''.format(tb=tb_source[1], sc=tb_source[0])
    type_list = exec_sql(exec_str, db=db)

    if False in [c in [c[0] for c in type_list] for c in list(new_col.keys()) + list(on_col.keys())]:
        raise ValueError('joinon: Either new_col or on_col are not in tb_source.')


    new_col = [tuple([vv]) + [t for t in type_list if kk in t][0]
               for kk, vv in new_col.items()]

    if new_columns:
        alt_str = ('ALTER TABLE ' + '.'.join(tb_target))
        add_str = ' ADD COLUMN IF NOT EXISTS '
        alt_str += ', '.join([add_str + ic + ' SMALLINT'
                              if type(ic) == str
                              else add_str + ' '.join([ic[i] for i in [0, 2]])
                              for ic in new_col]) + '; '
    else:
        alt_str = ''

    upd_str = 'UPDATE  ' + '.'.join(tb_target) + ' AS tg SET '
    new_cn = [ic if type(ic) == str else (ic[0], ic[1]) for ic in new_col]
    upd_str += ', '.join([ic[0] + ' = src.' + ic[1] for ic in new_cn])
    upd_str += ' FROM ' + '.'.join(tb_source) + ' AS src'
    upd_str += ' WHERE ' + ' AND '.join(['tg.' + ic[1] + ' = ' + 'src.' + ic[0]
                                      for ic in on_col.items()])
    upd_str += ';'

    exec_str = alt_str + upd_str

    if verbose:
        print(exec_str)

    time_msg = 'Join columns {} from {} to {}'.format(', '.join([c[0] for c in new_col]),
                                                      '.'.join(tb_source),
                                                      '.'.join(tb_target))
    exec_sql(exec_str, db=db, time_msg=time_msg)
    return exec_str

# %%
# CONSIDERED OBSOLETE DUE TO MISSING DB ARGUMENT

#def filtered_histogram(nbin=50,
#                       input_table = ['public', 'temp_mrg'],
#                       input_select = ['value', 'constr', 'encar', 'node',
#                                       'year', 'value_mask', 'fuel_mask',
#                                       'encar_mask', 'variable_mask'],
#                       input_filter = ['value_mask <> 0'],
#                       column_histogram = 'value',
#                       histogram_sum_cols = ['value', 'value_mask'],
#                       out_table=['out_st_lp_analysis', 'filtered_tr_prices']):
#
#    nbin=50
#    input_table = ['public', 'temp_mrg']
#    input_select = ['value', 'constr', 'encar', 'node',
#                    'year', 'value_mask', 'fuel_mask',
#                    'encar_mask', 'variable_mask']
#    input_filter = ['value_mask <> 0']
#    column_histogram = 'value'
#    histogram_sum_cols = ['value', 'value_mask']
#    out_table=['out_st_lp_analysis', 'filtered_tr_prices']
#
#    exec_str = ('DROP TABLE IF EXISTS ' + '.'.join(out_table) + '; ')
#    exec_str += (' WITH ' + input_table[1] + '_filt AS (' +
#                 ' SELECT ' + slct_str(input_select) +
#                 ' FROM ' + '.'.join(input_table) +
#                 ' ), ')
#    excl_list = [column_histogram] + histogram_sum_cols
#    exec_str += ('val_min_max AS ( SELECT *, ' +
#                 ' MIN(' + column_histogram + ') OVER (PARTITION BY ' +
#                 slct_str(input_select, exclude=excl_list) +
#                 ') as min, ' +
#                 ' MAX(' + column_histogram + ') OVER (PARTITION BY ' +
#                 slct_str(input_select, exclude=excl_list) +
#                 ') as max, ' +
#                 ' COUNT(' + column_histogram + ') OVER (PARTITION BY ' +
#                 slct_str(input_select, exclude=excl_list) +
#                 ') as cnt ' +
#                 ' FROM ' + input_table[1] + '_filt' +
#                 '), ')
#    exec_str += ('val_min_max_filt AS ( SELECT * FROM val_min_max' +
#                 ((' WHERE ' + ' AND '.join(input_filter))
#                   if input_filter != [] else '') +
#                 ')')
#    exec_str += (' SELECT ' +
#                 slct_str(input_select, exclude=excl_list) +
#                 ', MIN(min), MAX(max), ' +
#                 ' width_bucket(' + column_histogram +
#                                ', min, max * 1.0000000001, ' +
#                                str(nbin) + ') as bucket,' +
#                 ' COUNT(*) AS freq, ' +
#                 ', '.join(['SUM(' + c + ') AS ' + c + '_sum'
#                            for c in histogram_sum_cols]))
#    exec_str += (' INTO ' + '.'.join(out_table) +
#                 ' FROM val_min_max_filt ' +
#                 ' GROUP BY bucket '
#                 + slct_str(input_select, exclude=excl_list, commas=[',',''])
#                 )
#
#    conn = pg.connect(get_config('sql_connect')['psycopg2']
#                      .format(db=get_config('sql_connect')['db']))
#    cur = conn.cursor()
#    cur.execute(exec_str)
#    conn.commit()
#    try:
#        cur.fetchall()
#    except:
#        print("No results")
#    conn.close()
#
#    return dt.read_sql('storage1', *out_table), exec_str


# %% compare schemas

def compare_tables(db, list_sc, tb):
#
#    common_filt = [('swvr_vl', ['5.0%']),
#            ('swws_vl', ['All']),
#            ('swst_vl', ['0.0%']),
#            ('swtc_vl', ['CAS'])]
#
#    tb = 'par_fc_om'
#    list_sc = ['out_onlywinsol_new', 'out_nucspreadvr']

    flag_tb_exist = True

    list_tb_exist = []
    for nsc, isc in enumerate(list_sc):
        tb_exist = tb in get_sql_tables(isc, db)
        list_tb_exist.append((tb, isc, tb_exist))

        if not tb_exist:
            flag_tb_exist = False

    df_tot = pd.DataFrame()

    if flag_tb_exist:


        nsc, isc = list(enumerate(list_sc))[0]
        for nsc, isc in enumerate(list_sc):

        #    dict_ids =\
        #    {
        #     'nd_id': read_sql(db, isc, 'def_node').set_index('nd_id')['nd'].to_dict(),
        #     'pp_id': read_sql(db, isc, 'def_plant').set_index('pp_id')['pp'].to_dict(),
        #     }

#            loop_cols = get_sql_cols('def_loop', sc=isc, db=db)
#            slct_filt =  [f for f in common_filt if f[0] in loop_cols]
#            slct_filt = []

#            slct_run_id = read_sql(db, isc, 'def_loop', slct_filt).run_id.tolist()

#            df = read_sql(db, isc, tb, filt=[('run_id', slct_run_id)])
            df = read_sql(db, isc, tb)

            cols_index = ['pp_id', 'pp',
                          'nd_id', 'nd',
                          'fl_id', 'fl',
                          'pt_id', 'pt', 'pp_broad_cat',
                          'ca_id', 'ca',
                          'wk_id', 'wk',
                          'bool_out',
                          'hy', 'sy', 'run_id',
                          'mt_id', 'set_1_name', 'set_2_name', 'set_3_name',
                          'set_1_id', 'set_2_id', 'set_3_id', 'parameter',
                          'swhy_vl', 'color', 'mt']

#            cols = get_sql_cols(tb, sc=isc, db=db)
            slct_cols_index = [c for c in df.columns if c in cols_index]
            df = df.set_index(slct_cols_index).stack().rename(isc)

#            df = df.drop(['run_id'], axis=1).rename(columns={'value': 'value_' + str(nsc)})
#            df = df.rename(columns={'value': 'value_' + str(nsc)})

#            print(df)

            df_tot = pd.concat([df_tot, df], axis=1)


        df_tot = df_tot.fillna(0)

        df_tot['diff'] = (df_tot[list_sc[0]] - df_tot[list_sc[1]]).abs()
        df_tot['reldiff'] = df_tot['diff'] / df_tot[list_sc[0]]
        return df_tot.reset_index().sort_values(['reldiff'], ascending=False)

    else:
        print(list_tb_exist)
        return pd.DataFrame()


if __name__ == '__main__':

    list_sc = ['out_cal_1', 'out_cal_test_lin']
    db = 'storage2'
    tb = 'profdmnd'

    excl = ['profprice', 'profprice_soy', 'def_pp_type',
            'tm_soy_full', 'var_sy_pwr']

    excl += [tb for tb in get_sql_tables(list_sc[0], db) if 'analysis' in tb]
    excl += [tb for tb in get_sql_tables(list_sc[0], db) if 'var_sy' in tb]
    excl += [tb for tb in get_sql_tables(list_sc[0], db) if 'prof' in tb]

    for tb in [c for c in get_sql_tables(list_sc[0], db) if not c in excl]:
        print('*' * 30, tb, '*' * 30)
        dfcomp = compare_tables(db, list_sc, tb)
        dfcomp = dfcomp.loc[dfcomp['reldiff'] > 1e-5]#, ['nd_id', 'value_0', 'value_1', 'diff', 'reldiff']]


        if len(dfcomp) > 0:
            print(dfcomp)


# %%

def dump_by_table_sh(sc, db, target_dir):

    if __name__ == '__main__':
        sc='out_replace_basesmall'
        db='storage2'
        source_base='/run/user/1000/gvfs/dav:host=drive.switch.ch,ssl=true,prefix=%2Fremote.php%2Fdav/files/martin.soini@unige.ch'
        source_dir='SQL_DUMPS/out_replace_basesmall/'
        target_dir=os.path.join(source_base, source_dir)

    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    db_format = dict(user=None,
                     pw=None,
                     host=None,
                     port=None,
                     db=db)
    dbname = 'postgresql://{user}:{pw}@{host}:{port}/{db}'.format(**db_format)



    for itb in get_sql_tables(sc, db=db):
        tb = sc + '.' + itb
        fn = os.path.join(target_dir, itb + '.sql')
        print('Dumping table ', itb, ' to file ', fn)
        with open(fn, 'w+') as f:
            pg_dump('--dbname', dbname, '--table', tb, _out=f)

dbname = 'postgresql://postgres:postgres@localhost:5432/{db}'

def dump_by_table(sc, db, target_dir='C:\\Users\\ashreeta\\Documents\\Martin\\SWITCHdrive\\SQL_DUMPS\\out_disagg_new\\'):

#    if __name__ == '__main__':
#        sc='out_nucspreadvr_ext_it'
#        db='storage1'
#        target_dir='C:\\Users\\ashreeta\\Documents\\Martin\\SWITCHdrive\\SQL_DUMPS\\out_nucspreadvr_ext_it\\'

    exe = '\"C:\\Program Files\\PostgreSQL\\9.6\\bin\\pg_dump.exe\"'

    for itb in get_sql_tables(sc, db=db):
        tb = sc + '.' + itb
        fn = os.path.join(target_dir, tb + '.sql')
        print('Dumping table ', itb, ' to ', fn)
        run_str = ('{exe} --table {tb} --dbname={dbname} > {fn}'
                       .format(exe=exe, tb=tb, dbname=dbname.format(db=db), fn=fn))

        subprocess.run(run_str, shell=True, check=True)


def read_by_table(db, sc,
                  source_base='/run/user/1000/gvfs/dav:host=drive.switch.ch,ssl=true,prefix=%2Fremote.php%2Fdav/files/martin.soini@unige.ch',
                  source_dir='', warn_reset_schema=True,
                  patterns_only=False):

#    if __name__ == '__main__':
#        print('main')
#        db='storage2'
#        source_base='/run/user/1000/gvfs/dav:host=drive.switch.ch,ssl=true,prefix=%2Fremote.php%2Fdav/files/martin.soini@unige.ch'
#        source_dir='SQL_DUMPS/out_replace_vreseries/'


    source_dir = os.path.join(source_base, source_dir)


    f = []
    for (dirpath, dirnames, filenames) in walk(source_dir):
        f.extend(filenames)
        break

    if patterns_only:
        f = [fn for fn in f if any(pat in fn for pat in patterns_only)]


    print('Reading from %s, %d files found.' % (source_dir, len(f)))

    exe = 'psql'

    reset_schema(sc, db, warn=warn_reset_schema)

    db_format = dict(user=None,
                     pw=None,
                     host=None,
                     port=None,
                     db=db)
    dbname = 'postgresql://{user}:{pw}@{host}:{port}/{db}'.format(**db_format)

    for file in f:
        print('Reading file ', file)
        fn = source_dir + file

        run_str = '{exe} --dbname={dbname} < {fn}'.format(exe=exe, dbname=dbname, fn=fn)
        subprocess.run(run_str, shell=True, check=True)

if __name__ == '__main__':
    pass

#    source_dir = 'SQL_DUMPS/out_replace_emission/'
#    sc = 'out_replace_emission'
#    read_by_table(db, sc, source_dir=source_dir, warn_reset_schema=False,
#                  patterns_only=['analysis_', 'def_', 'encar'])

#    dump_by_table('out_marg_store', 'storage2', target_dir='C:\\Users\\ashreeta\\Documents\\Martin\\SWITCHdrive\\SQL_DUMPS\\out_marg_store_new\\')

# %%






import fnmatch
from setuptools import find_packages, setup
from setuptools.command.build_py import build_py as build_py_orig

with open("description_pypi.rst", "r") as fh:
    long_description = fh.read()

setup(
    name="Grimsel",
    version="0.0.6",
    author="Grimsel contributors listed in AUTHORS",
    author_email="mcsoini_dev@posteo.org",
    description=("GeneRal Integrated Modeling environment for the Supply of"
                 " Electricity and Low-temperature heat"),
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/mcsoini/grimsel",
    packages=find_packages(),
    install_requires=['pyutilib>=5.8.0',
                      'pandas>=1.0.0', 
                      'pyomo==5.6.9',
                      'wrapt>=1.12.1', 
                      'psycopg2>=2.8.5', 
                      'numpy>=1.18.2', 
                      'sqlalchemy>=1.3.16',
                      'statsmodels>=0.11.1',
                      'tables>=3.6.1',
                      'fastparquet>=0.3.3',
                      'matplotlib>=3.2.1',
                      'tabulate>=0.8.7'
                      ],
     classifiers=[
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent"],
)





import logging

def _get_logger(name):
    logger = logging.getLogger(name)

    if not logger.hasHandlers():
        f_handler = logging.StreamHandler()
        f_handler.setLevel(0)
        format_str = '> %(asctime)s - %(levelname)s - %(name)s - %(message)s'
        f_format = logging.Formatter(format_str, "%H:%M:%S")
        f_handler.setFormatter(f_format)
        logger.addHandler(f_handler)

    return logger

logger = _get_logger(__name__)

from grimsel.core.model_base import ModelBase
from grimsel.core.model_loop import ModelLoop

from grimsel.auxiliary.maps import Maps

from grimsel.auxiliary.multiproc import run_parallel, run_sequential




'''
Model variables
=================


'''

import pyomo.environ as po
from collections import namedtuple

from grimsel import _get_logger
from grimsel.auxiliary.aux_general import silence_pd_warning

logger = _get_logger(__name__)


VAR_DOCS = {'pwr': ':math:`p_\\mathrm{t,p,c} \\forall sy\\_ppall\\_ca \\in (0,\\infty)`: per time slot production of energy carriers :math:`\\mathrm{ca}` from all plants',
         'pwr_ramp': ':math:`\\delta p_\\mathrm{t,p,c} \\forall sy\\_rp\\_ca \\in (-\\infty,\\infty)`: per time slot ramping power difference for relevant plants :math:`\\mathrm{rp}`',
#         'pwr_ramp_abs': ':math:`|\\delta p_\\mathrm{t,p,c}| \\forall sy\\_rp\\_ca \\in (0,\\infty)`: per time slot absolute ramping power difference',
         'pwr_ramp_abs': '|pwr_ramp_abs| `\\forall sy\\_rp\\_ca \\in (0,\\infty)`: per time slot absolute ramping power difference',
         'pwr_st_ch': ':math:`p_\\mathrm{chg,t,p,c} \\forall sy\\_st\\_ca \\in (0,\\infty)`: per time slot charging power of storage plants',
         'erg_st': ':math:`e_\\mathrm{t,p,c} \\forall sy\\_st\\_ca\\cup sy\\_hyrs\\_ca \\in (0,\\infty)`: stored energy in storage and reservoirs each time slot',
         'trm': ':math:`p_\\mathrm{trm,t,n,n_2,c} \\forall symin\\_ndcnn \\in (-\\infty,\\infty)`: internodal power transmission for each of the time slots',
         'erg_mt': ':math:`E_\\mathrm{m,p,c} \\forall mt \\times hyrs\\_ca\\in (0,\\infty)`: monthly produced energy from hydro reservoirs',
         'erg_fl_yr': ':math:`E_\\mathrm{p,n,c,f} \\forall ppall\\_ndcafl\\in (0,\\infty)`: yearly produced energy by plant and fuel',
         'erg_yr': ':math:`E_\\mathrm{p,c} \\forall ppall\\_ca \\in (0,\\infty)`: yearly produced energy by plant',
         'pwr_ramp_yr': ':math:`\\Delta p_\\mathrm{p,c} \\forall rp\\_ca \\in (0,\\infty)`: yearly aggregated absolute ramping',
         'vc_fl_pp_yr': ':math:`{c}_\\mathrm{fuel, p,c,f} \\forall ppall\\_cafl\\setminus lin\\_cafl \\in (-\\infty,\\infty)`: yearly variable fuel cost (only constant supply curves)',
         'vc_om_pp_yr': ':math:`{c}_\\mathrm{om_v,p,c} \\forall ppall\\_ca \\in (0,\\infty)`: yearly variable O\\&M cost',
         'fc_om_pp_yr': ':math:`{c}_\\mathrm{om_f, p,c} \\forall ppall\\_ca \\in (0,\\infty)`: yearly fixed O\\&M cost',
         'fc_cp_pp_yr': ':math:`{c}_\\mathrm{cp, p,c} \\forall add\\_ca \\in (0,\\infty)`: yearly capital investment cost',
         'vc_co2_pp_yr': ':math:`{c}_\\mathrm{em, p,c} \\forall pp\\_ca \\in (0,\\infty)`: yearly |CO2| emission cost',
         'vc_ramp_yr': ':math:`{c}_\\mathrm{rp, p,c} \\forall rp\\_ca \\in (0,\\infty)`: yearly ramping cost',
         'cap_pwr_tot': ':math:`P_\\mathrm{tot, p, c} \\forall ppall\\_ca \\in (0,\\infty)`: net installed power capacity',
         'cap_pwr_new': ':math:`P_\\mathrm{new, p,c} \\forall add\\_ca \\in (0,\\infty)`: new installed power capacity',
         'cap_pwr_rem': ':math:`P_\\mathrm{ret, p,c} \\forall rem\\_ca \\in (0,\\infty)`: retired power capacity',
         'cap_erg_tot': ':math:`C_\\mathrm{ret, p,c} \\forall rem\\_ca \\in (0,\\infty)`: total energy capacity ofr storage and reservoirs'}

class Variables:
    '''
    Mixin class containing all variable definitions.
    '''





    def define_variables(self):
        r'''
        Adds all variables to the model instance by calling :func:`vadd`.

        '''

        Var = namedtuple('Var', ['name', 'sets', 'bounds'])

        vars_ = [Var('pwr', self.sy_ppall_ca, None),
                 Var('pwr_ramp', self.sy_rp_ca, (None, None)),
                 Var('pwr_ramp_abs', self.sy_rp_ca, None),
                 Var('pwr_st_ch', self.sy_st_ca, None),
                 Var('erg_st', self.sy_st_ca | self.sy_hyrs_ca, None),

                 Var('trm', self.symin_ndcnn, (None, None)),
                 Var('erg_mt', (self.mt, self.hyrs_ca), None),
                 Var('erg_fl_yr', self.ppall_ndcafl, None),
                 Var('erg_yr', self.ppall_ca, None),
                 Var('pwr_ramp_yr', self.rp_ca, None),

                 Var('vc_fl_pp_yr', self.ppall_cafl - self.lin_cafl, (None, None)),
                 Var('vc_om_pp_yr', self.ppall_ca, None),
                 Var('fc_om_pp_yr', self.ppall_ca, None),
                 Var('fc_cp_pp_yr', self.add_ca, None),
                 Var('vc_co2_pp_yr', self.pp_ca, None),
                 Var('vc_ramp_yr', self.rp_ca, None),
                 Var('cap_pwr_tot', self.ppall_ca, None),
                 Var('cap_pwr_new', self.add_ca, None),
                 Var('cap_pwr_rem', self.rem_ca, None),
                 Var('cap_erg_tot', self.st_ca | self.hyrs_ca, None),
                 ]

        for var in vars_:
            self.delete_component(var.name)
            self.vadd(var.name, var.sets,
                      var.bounds if var.bounds else (0, None),
                      doc=VAR_DOCS[var.name])



    def vadd(self, variable_name, variable_index, bounds=(0, None),
             domain=po.Reals, doc=''):
        '''
        Adds a single variable to the model instance.

        Makes sure the ``variable_index`` is a tuple of pyomo sets and
        checks whether the sets are None or empty

        Parameters
        ----------
        variable_name : str
            name of the variable attribute
        variable_index : pyomo set
            parameter index set
        bounds : tuple
            upper and lower bounds of the variable defined through len-2 tuple;
            default: ``(0, None)``
        domain : pyomo domain
            default: ``pyomo.environ.Reals``
        doc : str
            parameter docstring; default ''

        '''

        if not type(variable_index) is tuple:
            variable_index = (variable_index,)

        logger.info('Defining variable %s ...'%variable_name)

        if not self.check_valid_indices(variable_index):
            return None
        else:
            logger.info('... ok.')

        setattr(self, variable_name, po.Var(*variable_index, bounds=bounds,
                                            domain=domain, doc=doc))


    @silence_pd_warning
    @staticmethod
    def _get_set_docs():

        import pandas as pd
        import tabulate


        var_doc = pd.Series(VAR_DOCS).apply(lambda x: r'{}'.format(x))
        var_doc = var_doc.reset_index()
        var_doc.columns = ['Attribute', 'Symbol']

        def get_doc_row(x):
            x_spl = x.Symbol.split('\in ')
            x_spl2 = x_spl[1].split('infty)`:')
            x_spl3 = x_spl[0].split(r'\forall')

            dct_row = {'Attribute': '``%s``'%x.Attribute,
                       'Symbol': x_spl3[0].strip(' ') + '`',
                       'Set': ':math:`\mathrm{' + x_spl3[1].strip(' ') + '}`',
                       'Bounds': ':math:`' + x_spl2[0] + 'infty)`',
                       'Doc': x_spl2[1].strip(' ').strip('\n')}

            return pd.Series(dct_row)

        var_doc = var_doc.apply(get_doc_row, axis=1)


        return tabulate.tabulate(var_doc, tablefmt='rst', showindex=False,
                          headers=var_doc.columns)


Variables.define_variables.__doc__ += '\n'*3 + Variables._get_set_docs()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov  5 21:31:46 2019

@author: user
"""
from multiprocessing import Pool
from multiprocessing import current_process
import contextlib
from grimsel.core.model_loop import logger_parallel
from grimsel import logger

def _call_list_run_id(func, list_run_id):

    for run_id in list_run_id:
        func(run_id)



@contextlib.contextmanager
def _adjust_logger_levels(do, ml, grimsel_level, parallel_level, verbose_solver):

    _verbose_solver = ml.m.verbose_solver
    old_grimsel_level = grimsel_level
    old_parallel_level = parallel_level

    if do:
        ml.m.verbose_solver = verbose_solver
        logger.setLevel(grimsel_level)
        logger_parallel.setLevel(parallel_level)

    yield

    ml.m.verbose_solver = _verbose_solver
    logger.setLevel(old_grimsel_level)
    logger_parallel.setLevel(old_parallel_level)


def run_sequential(ml, func, adjust_logger_levels=True):
    ''' Sequential execution of all model runs. '''

    with _adjust_logger_levels(adjust_logger_levels,
                               ml, 'DEBUG', 'ERROR', True):

        _call_list_run_id(func, ml.get_list_run_id())



def run_parallel(ml, func, nproc=None, groupby=None,
                 adjust_logger_levels=True):
    '''
    Parameters
    ----------
    func : function(run_id)
        function to be sent to the workers
    nproc : int
        Number of processes
    groupby : list of `ModelLoop.df_run` columns
        Determines the groups of runs which are passed to the processes. This
        is necessary if certain model runs depend on each other.
    '''

    with _adjust_logger_levels(adjust_logger_levels,
                               ml, 'ERROR', 'INFO', False):

        p = Pool(nproc)

        if groupby:
            # list of lists of run_ids grouped by groupby
            grouped_run_id = (ml.df_def_run.groupby(groupby)
                                           .run_id.apply(list).tolist())

            args = zip([func] * len(grouped_run_id), grouped_run_id)
            p.starmap(_call_list_run_id, args)

        else:
            list_run_id = ml.get_list_run_id()
            p.map(func, list_run_id)

        p.close()
        p.join()

        ml._merge_df_run_files()




